import math
from functools import partial

import torch.nn as nn

from bcos.data.presets import (
    CIFAR10ClassificationPresetTest,
    CIFAR10ClassificationPresetTrain,
)
from bcos.experiments.utils import (
    configs_cli,
    create_configs_with_different_seeds,
    update_config,
)
from bcos.modules import norms
from bcos.modules.losses import BinaryCrossEntropyLoss
from bcos.optim import LRSchedulerFactory, OptimizerFactory

__all__ = ["CONFIGS"]

NUM_CLASSES = 10

# This config is based on
# https://github.com/moboehle/B-cos/blob/main/experiments/CIFAR10/bcos/experiment_parameters.py
DEFAULT_NUM_EPOCHS = 100
DEFAULT_LR = 1e-3
DEFAULT_BATCH_SIZE = 64

DEFAULTS = dict(
    data=dict(
        train_transform=CIFAR10ClassificationPresetTrain(is_bcos=True),
        test_transform=CIFAR10ClassificationPresetTest(is_bcos=True),
        batch_size=DEFAULT_BATCH_SIZE,
        num_workers=4,
        num_classes=NUM_CLASSES,
    ),
    model=dict(
        is_bcos=True,
        # "name": None,
        args=dict(
            num_classes=NUM_CLASSES,
        ),
        bcos_args=dict(
            b=2,
        ),
    ),
    criterion=BinaryCrossEntropyLoss(),
    test_criterion=BinaryCrossEntropyLoss(),
    optimizer=OptimizerFactory(name="Adam", lr=DEFAULT_LR),
    lr_scheduler=LRSchedulerFactory(
        name="cosineannealinglr", epochs=DEFAULT_NUM_EPOCHS
    ),
    trainer=dict(
        max_epochs=DEFAULT_NUM_EPOCHS,
    ),
)


# helper
def update_default(new_config):
    return update_config(DEFAULTS, new_config)


RESNET_DEPTHS = [20, 32, 44, 56]
NORMS_MAP = {
    # centered
    "an": norms.centered_norms.AllNorm2d,
    "bn": norms.centered_norms.BatchNorm2d,
    "gn": partial(norms.centered_norms.DetachableGroupNorm2d, 2),  # num_groups
    "in": norms.centered_norms.DetachableGNInstanceNorm2d,
    "ln": norms.centered_norms.DetachableGNLayerNorm2d,
    "pn": norms.centered_norms.DetachablePositionNorm2d,
    # uncentered
    "anu": norms.uncentered_norms.AllNormUncentered2d,
    "bnu": norms.uncentered_norms.BatchNormUncentered2d,
    "gnu": partial(norms.uncentered_norms.GroupNormUncentered2d, 2),
    "inu": norms.uncentered_norms.GNInstanceNormUncentered2d,
    "lnu": norms.uncentered_norms.GNLayerNormUncentered2d,
    "pnu": norms.uncentered_norms.PositionNormUncentered2d,
}

resnets = {
    f"resnet_{d}": update_default(
        dict(
            model=dict(
                name=f"resnet{d}",
                args=dict(
                    norm_layer=nn.Identity,
                    logit_bias=-math.log(NUM_CLASSES - 1),
                ),
                bcos_args=dict(
                    max_out=2,
                ),
            ),
        )
    )
    for d in RESNET_DEPTHS
}

resnets_norms = {
    f"{name}_{norm_name}{affine_or_not_suffix}": update_config(
        old_c,
        dict(
            model=dict(
                args=dict(
                    norm_layer=affine_or_not_wrapper(norm_layer),
                ),
            ),
        ),
    )
    for name, old_c in resnets.items()
    for norm_name, norm_layer in NORMS_MAP.items()
    for affine_or_not_suffix, affine_or_not_wrapper in [
        ("", lambda x: x),
        ("-linear", norms.NoBias),
        ("-unaffine", norms.Unaffine),
    ]
}

resnets_nomaxout = {
    f"{name}-nomaxout": update_config(
        old_c,
        dict(
            model=dict(
                bcos_args=dict(
                    max_out=1,
                )
            )
        ),
    )
    for config_group in [resnets, resnets_norms]
    for name, old_c in config_group.items()
}


CONFIGS = dict()
CONFIGS.update(resnets)
CONFIGS.update(resnets_norms)
CONFIGS.update(resnets_nomaxout)
CONFIGS.update(create_configs_with_different_seeds(CONFIGS, seeds=[420, 1337]))


if __name__ == "__main__":
    configs_cli(CONFIGS)



"""
Some extra utilities.
"""
from functools import wraps

__all__ = ["NoBias", "Unaffine"]


def _append_to_name(mod, suffix):
    old_name = mod.__class__.__name__

    def new_name():
        return old_name + suffix

    mod._get_name = new_name


def NoBias(make_layer):
    """
    Wraps around the layer making function and removes bias by setting it to
    None after instantiation.

    Parameters
    ----------
    make_layer
        The layer making function.

    Returns
    -------
    Modified layer making function which sets bias to None.
    """

    @wraps(make_layer)
    def init(*args, **kwargs):
        norm = make_layer(*args, **kwargs)

        assert (
            norm.bias is not None
        ), "It makes no sense to use this wrapper if you set affine=False!"
        norm.bias = None

        _append_to_name(norm, "NoBias")

        return norm

    if hasattr(make_layer, "__name__"):
        init.__name__ = make_layer.__name__ + "NoBias"
    if hasattr(make_layer, "__qualname__"):
        init.__qualname__ = make_layer.__qualname__ + "NoBias"

    return init


def Unaffine(make_layer):
    """
    Wraps around the layer making function and removes bias and weight
    by setting them to None after instantiation.

    Parameters
    ----------
    make_layer
        The layer making function.

    Returns
    -------
    Modified layer making function which sets bias and weight to None.
    """

    @wraps(make_layer)
    def init(*args, **kwargs):
        norm = make_layer(*args, **kwargs)

        assert (
            norm.bias is not None
        ), "It makes no sense to use this wrapper if you set affine=False!"
        norm.bias = None
        norm.weight = None

        _append_to_name(norm, "Unaffine")

        return norm

    if hasattr(make_layer, "__name__"):
        init.__name__ = make_layer.__name__ + "Unaffine"
    if hasattr(make_layer, "__qualname__"):
        init.__qualname__ = make_layer.__qualname__ + "Unaffine"

    return init



from torch import nn

from bcos.common import BcosUtilMixin

__all__ = ["DetachableModule", "BcosSequential"]


class DetachableModule(nn.Module):
    """
    A base module for modules which can detach dynamic weights from the graph,
    which is necessary to calculate explanations.
    """

    def __init__(self):
        super().__init__()
        self.detach = False

    def set_explanation_mode(self, activate: bool = True) -> None:
        """
        Turn explanation mode on or off.

        Parameters
        ----------
        activate : bool
            Turn it on.
        """
        self.detach = activate

    @property
    def is_in_explanation_mode(self) -> bool:
        """
        Whether the module is in explanation mode or not.
        """
        return self.detach


class BcosSequential(BcosUtilMixin, nn.Sequential):
    """
    Wrapper for models which are nn.Sequential at the "root" module level.
    This only adds helper functionality from `BcosMixIn`.
    """

    def __init__(self, *args):
        super().__init__(*args)



"""
All norm without centering.

Code partially taken from
https://github.com/pytorch/pytorch/blob/9e81c0c3f46a36333e82b799b4afa79b44b6bb59/torch/nn/modules/batchnorm.py
"""
from typing import Optional

import torch.nn as nn
from torch import Tensor

from bcos.modules.common import DetachableModule

__all__ = [
    "AllNormUncentered2d",
]


def all_norm_uncentered_2d(
    input: Tensor,
    running_var: Optional[Tensor],
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    training: bool = False,
    momentum: float = 0.1,
    eps: float = 1e-5,
    detach: bool = False,
):
    """
    Uncentered AllNorm. Accepts only batched color image tensors.
    """
    assert input.dim() == 4, "input should be a 4d tensor!"

    if training:
        # first calc stats
        x = input.detach() if detach else input
        var = x.var(unbiased=False)

        # update running stats if given
        if running_var is not None:
            running_var.copy_((1 - momentum) * running_var + momentum * var.detach())

    else:  # evaluation mode
        assert running_var is not None, "running_var must be defined in eval mode"
        var = running_var

    std = (var + eps).sqrt()[None, ..., None, None]

    result = input / std

    if weight is not None:
        result = weight[None, ..., None, None] * result
    if bias is not None:
        result = result + bias[None, ..., None, None]

    return result


class AllNormUncentered2d(nn.BatchNorm2d, DetachableModule):
    def __init__(
        self,
        num_features: int,
        *args,
        **kwargs,
    ) -> None:
        DetachableModule.__init__(self)

        # since we do it over the whole thing we have to set
        # this to one
        super().__init__(
            1,
            *args,
            **kwargs,
        )

    def forward(self, input):
        self._check_input_dim(input)  # require 4

        # exponential_average_factor is set to self.momentum
        # (when it is available) only so that it gets updated
        # in ONNX graph when this node is exported to ONNX.
        if self.momentum is None:
            exponential_average_factor = 0.0
        else:
            exponential_average_factor = self.momentum

        if self.training and self.track_running_stats:
            # TODO: if statement only here to tell the jit to skip emitting this when it is None
            if self.num_batches_tracked is not None:  # type: ignore[has-type]
                self.num_batches_tracked.add_(1)  # type: ignore[has-type]
                if self.momentum is None:  # use cumulative moving average
                    exponential_average_factor = 1.0 / float(self.num_batches_tracked)
                else:  # use exponential moving average
                    exponential_average_factor = self.momentum

        r"""
        Decide whether the mini-batch stats should be used for normalization rather than the buffers.
        Mini-batch stats are used in training mode, and in eval mode when buffers are None.
        """
        if self.training:
            bn_training = True
        else:
            bn_training = (self.running_mean is None) and (self.running_var is None)

        r"""
        Buffers are only updated if they are to be tracked and we are in training mode. Thus they only need to be
        passed when the update should occur (i.e. in training mode when they are tracked), or when buffer stats are
        used for normalization (i.e. in eval mode when buffers are not None).
        """
        return all_norm_uncentered_2d(
            input=input,
            # If buffers are not to be tracked, ensure that they won't be updated
            running_var=self.running_var
            if not self.training or self.track_running_stats
            else None,
            weight=self.weight,
            bias=self.bias,
            training=bn_training,
            momentum=exponential_average_factor,
            eps=self.eps,
            detach=self.detach,
        )



"""
Nothing fancy, just for creating lr schedulers.

Modified from: https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf054/references/classification/train.py
"""
from typing import Literal, Optional

import torch

__all__ = ["LRSchedulerFactory"]


class LRSchedulerFactory:
    def __init__(
        self,
        name: str,
        step_size: int = 30,  # interpreted as epochs, converted to steps if needed
        gamma: float = 0.1,
        epochs: Optional[int] = None,
        lr_min: float = 0.0,
        interval: Literal["epoch", "step"] = "epoch",
        # warmup
        warmup_epochs: Optional[int] = None,
        warmup_steps: Optional[int] = None,
        warmup_method: str = "constant",
        warmup_decay: float = 0.01,
    ):
        self.name = name
        self.step_size = step_size
        self.gamma = gamma
        self.epochs = epochs
        self.lr_min = lr_min

        self.interval = interval

        self.warmup_epochs = warmup_epochs
        self.warmup_steps = warmup_steps
        self.warmup_method = warmup_method
        self.warmup_decay = warmup_decay

        assert name.lower() in ["steplr", "cosineannealinglr", "exponentiallr"]
        assert warmup_method.lower() in ["linear", "constant"]
        assert interval in ["epoch", "step"]
        assert name.lower() != "cosineannealinglr" or (
            self.epochs is not None and self.epochs > 0
        ), "If name='cosineannealinglr' then (positive) epochs need to be provided!"
        assert interval == "epoch" or (
            self.epochs is not None and self.epochs > 0
        ), "If interval='step' then (positive) epochs need to be provided!"
        assert (
            warmup_steps is None or warmup_epochs is None
        ), "Do not provide both warmup_steps and warmup_epochs at the same time!"
        assert (
            interval == "step" or self.warmup_steps is None
        ), "For warmup_steps, interval must be 'step'"
        assert (
            warmup_epochs is None or warmup_epochs > 0
        ), "Provided positive warmup_epochs, if then"
        assert (
            warmup_steps is None or warmup_steps > 0
        ), "Provided positive warmup_steps, if then"

    def create(
        self, optimizer: torch.optim.Optimizer, total_steps: Optional[int] = None
    ):
        assert (
            total_steps is not None or self.interval == "epoch"
        ), "interval='step' requires total_steps!"

        name = self.name.lower()
        if name == "steplr":
            step_size = self.step_size
            if self.interval == "step":
                num_steps_per_epoch = total_steps // self.epochs
                step_size *= num_steps_per_epoch
            main_lr_scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer, step_size=step_size, gamma=self.gamma
            )
        elif name == "cosineannealinglr":
            T_max = self.epochs
            if self.warmup_epochs is not None:
                T_max -= self.warmup_epochs
            if self.interval == "step":
                num_steps_per_epoch = total_steps // self.epochs
                T_max *= num_steps_per_epoch
                if self.warmup_steps is not None:
                    assert self.warmup_epochs is None
                    T_max -= self.warmup_steps

            main_lr_scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                optimizer, T_max=T_max, eta_min=self.lr_min
            )
        elif name == "exponentiallr":
            main_lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(
                optimizer, gamma=self.gamma
            )
        else:
            raise RuntimeError(
                f"Invalid lr scheduler '{name}'. Only StepLR, CosineAnnealingLR and ExponentialLR "
                "are supported."
            )

        if self.warmup_epochs or self.warmup_steps:
            warmup_method = self.warmup_method.lower()
            if self.interval == "epoch":
                total_iters = self.warmup_epochs
            else:
                num_steps_per_epoch = total_steps // self.epochs
                if self.warmup_steps is not None:
                    total_iters = self.warmup_steps
                else:
                    total_iters = self.warmup_epochs * num_steps_per_epoch

            if warmup_method == "linear":
                warmup_lr_scheduler = torch.optim.lr_scheduler.LinearLR(
                    optimizer,
                    start_factor=self.warmup_decay,
                    total_iters=total_iters,
                )
            elif warmup_method == "constant":
                warmup_lr_scheduler = torch.optim.lr_scheduler.ConstantLR(
                    optimizer,
                    factor=self.warmup_decay,
                    total_iters=total_iters,
                )
            else:
                raise ValueError(
                    f"Invalid warmup lr method '{self.warmup_method}'. Only linear and constant are supported."
                )
            lr_scheduler = torch.optim.lr_scheduler.SequentialLR(
                optimizer,
                schedulers=[warmup_lr_scheduler, main_lr_scheduler],
                milestones=[total_iters],
            )
        else:
            lr_scheduler = main_lr_scheduler

        if self.interval == "epoch":
            return lr_scheduler

        return dict(
            scheduler=lr_scheduler,
            interval="step",
        )

    def with_epochs(self, epochs: Optional[int]):
        return type(self)(
            name=self.name,
            step_size=self.step_size,
            gamma=self.step_size,
            epochs=epochs,
            lr_min=self.lr_min,
            interval=self.interval,
            warmup_epochs=self.warmup_epochs,
            warmup_decay=self.warmup_decay,
            warmup_method=self.warmup_method,
        )

    def __repr__(self):
        s = self.__class__.__name__
        name = self.name.lower()
        s += f"(scheduler='{name}'"

        args = dict()
        if name == "steplr":
            args = dict(step_size=self.step_size, gamma=self.gamma)
        elif name == "cosineannealinglr":
            args = dict(epochs=self.epochs, lr_min=self.lr_min)
        elif name == "exponentiallr":
            args = dict(gamma=self.gamma)

        if self.warmup_epochs or self.warmup_steps:
            args.update(
                dict(
                    warmup_method=self.warmup_method,
                    warmup_epochs=self.warmup_epochs,
                    warmup_steps=self.warmup_steps,
                    warmup_decay=self.warmup_decay,
                )
            )

        args.update(dict(interval=self.interval))

        if args:
            for k, v in args.items():
                s += f", {k}={v}"
        s += ")"

        return s

    def __to_config__(self):
        """See bcos.experiments.utils.sanitize_config for details."""
        name = self.name.lower()
        result = dict(name=name)

        if name == "steplr":
            result.update(dict(step_size=self.step_size, gamma=self.gamma))
        elif name == "cosineannealinglr":
            result.update(dict(epochs=self.epochs, lr_min=self.lr_min))
        elif name == "exponentiallr":
            result.update(dict(gamma=self.gamma))

        if self.warmup_epochs or self.warmup_steps:
            result.update(
                dict(
                    warmup_method=self.warmup_method,
                    warmup_epochs=self.warmup_epochs,
                    warmup_steps=self.warmup_steps,
                    warmup_decay=self.warmup_decay,
                )
            )

        result.update(dict(interval=self.interval))

        return result



__all__ = [
    "AmbiguityError",
    "EMANotFound",
    "MetricsNotFoundError",
    "ModelFactoryDoesNotSupportPretrainedError",
]


class EMANotFound(Exception):
    pass


class MetricsNotFoundError(Exception):
    pass


class ModelFactoryDoesNotSupportPretrainedError(Exception):
    pass


class AmbiguityError(Exception):
    pass



from . import norms
from .bcosconv2d import BcosConv2d, BcosConv2dWithScale
from .bcoslinear import BcosLinear
from .common import BcosSequential, DetachableModule
from .logitlayer import LogitLayer
from .losses import *
from .norms import *



"""
Caching into SHM.
"""
import os
from pathlib import Path

import bcos.settings as settings


def cache_tar_files_to_shm() -> None:
    """
    Caches imagenet tar files into shm.
    Intended to work only for imagenet.
    """
    IMAGENET_TRAIN_TAR_FILES_DIR = settings.IMAGENET_TRAIN_TAR_FILES_DIR
    SHMTMPDIR = settings.SHMTMPDIR
    SHMTMPDIR = Path(SHMTMPDIR)
    TRAIN_DIRECTORY = SHMTMPDIR / "train"

    tar_files_to_extract = list(Path(IMAGENET_TRAIN_TAR_FILES_DIR).iterdir())
    assert len(tar_files_to_extract) == 1_000

    TRAIN_DIRECTORY.mkdir(exist_ok=True)

    os.system(f"df -H {SHMTMPDIR}")
    os.system(
        f"echo {' '.join(map(str, tar_files_to_extract))} "
        f"| tr ' ' '\n' "
        f"| xargs -P {os.cpu_count()} "
        "-I {} cp {} "
        f"{SHMTMPDIR}"
    )
    os.system(
        f"echo {' '.join([str(SHMTMPDIR / p.name) for p in tar_files_to_extract])} "
        f"| tr ' ' '\n' "
        f"| xargs -P {os.cpu_count()} "
        "-I {} tar xf {}"
        f" --directory {TRAIN_DIRECTORY}"
    )
    os.system(
        f"echo {' '.join([str(SHMTMPDIR / p.name) for p in tar_files_to_extract])} "
        f"| tr ' ' '\n' "
        f"| xargs -P {os.cpu_count()} -n 8 rm -rf"
    )
    os.system(f"df -H {SHMTMPDIR}")



import torch
from torchvision.transforms import autoaugment, transforms
from torchvision.transforms.functional import InterpolationMode

import bcos.data.transforms as custom_transforms

IMAGENET_MEAN = (0.485, 0.456, 0.406)
IMAGENET_STD = (0.229, 0.224, 0.225)


class ImageNetClassificationPresetTrain:
    def __init__(
        self,
        *,
        crop_size,
        mean=IMAGENET_MEAN,
        std=IMAGENET_STD,
        interpolation=InterpolationMode.BILINEAR,
        hflip_prob=0.5,
        auto_augment_policy=None,
        ra_magnitude=9,
        augmix_severity=3,
        random_erase_prob=0.0,
        is_bcos=False,
    ):
        self.args = get_args_dict(ignore=["mean", "std"])
        trans = [transforms.RandomResizedCrop(crop_size, interpolation=interpolation)]
        if hflip_prob > 0:
            trans.append(transforms.RandomHorizontalFlip(hflip_prob))
        if auto_augment_policy is not None:
            if auto_augment_policy == "ra":
                trans.append(
                    autoaugment.RandAugment(
                        interpolation=interpolation, magnitude=ra_magnitude
                    )
                )
            elif auto_augment_policy == "ta_wide":
                trans.append(
                    autoaugment.TrivialAugmentWide(interpolation=interpolation)
                )
            elif auto_augment_policy == "augmix":
                trans.append(
                    autoaugment.AugMix(
                        interpolation=interpolation, severity=augmix_severity
                    )
                )
            else:
                aa_policy = autoaugment.AutoAugmentPolicy(auto_augment_policy)
                trans.append(
                    autoaugment.AutoAugment(
                        policy=aa_policy, interpolation=interpolation
                    )
                )
        trans.extend(
            [
                transforms.PILToTensor(),
                transforms.ConvertImageDtype(torch.float),
            ]
        )
        if not is_bcos:
            trans.append(transforms.Normalize(mean=mean, std=std))
        if random_erase_prob > 0:
            trans.append(transforms.RandomErasing(p=random_erase_prob))

        if is_bcos:
            trans.append(custom_transforms.AddInverse())

        self.transforms = transforms.Compose(trans)

    def __call__(self, img):
        return self.transforms(img)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.transforms.transforms)})"

    def __rich_repr__(self):
        # https://rich.readthedocs.io/en/stable/pretty.html
        yield "transforms", self.transforms.transforms

    def __to_config__(self):
        """See bcos.experiments.utils.sanitize_config for details."""
        result = dict(
            transform=repr(self),
            **self.args,
        )
        result["interpolation"] = str(result["interpolation"])
        return result


class ImageNetClassificationPresetEval:
    def __init__(
        self,
        *,
        crop_size,
        resize_size=256,
        mean=IMAGENET_MEAN,
        std=IMAGENET_STD,
        interpolation=InterpolationMode.BILINEAR,
        is_bcos=False,
    ):
        self.args = get_args_dict(ignore=["mean", "std"])
        trans = [
            transforms.Resize(resize_size, interpolation=interpolation),
            transforms.CenterCrop(crop_size),
            transforms.PILToTensor(),
            transforms.ConvertImageDtype(torch.float),
            custom_transforms.AddInverse()
            if is_bcos
            else transforms.Normalize(mean=mean, std=std),
        ]

        self.transforms = transforms.Compose(trans)

    def __call__(self, img):
        return self.transforms(img)

    @property
    def resize(self):
        return self.transforms.transforms[0]

    @property
    def center_crop(self):
        return self.transforms.transforms[1]

    def no_scale(self, img):
        x = img
        for t in self.transforms.transforms[2:]:
            x = t(x)
        return x

    # this is intended for when using the pretrained models
    def transform_with_options(self, img, center_crop=True, resize=True):
        x = img
        if resize:
            x = self.resize(x)
        if center_crop:
            x = self.center_crop(x)
        x = self.no_scale(x)
        return x

    def with_args(self, **kwargs):
        args = self.args.copy()
        args.update(kwargs)
        return self.__class__(**args)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.transforms.transforms)})"

    def __rich_repr__(self):
        yield "transforms", self.transforms.transforms

    def __to_config__(self):
        """See bcos.experiments.utils.sanitize_config for details."""
        result = dict(
            transform=repr(self),
            **self.args,
        )
        result["interpolation"] = str(result["interpolation"])
        return result


CIFAR10_MEAN = (0.49139968, 0.48215841, 0.44653091)
CIFAR10_STD = (0.24703223, 0.24348513, 0.26158784)


class CIFAR10ClassificationPresetTrain:
    def __init__(
        self,
        *,
        mean=CIFAR10_MEAN,
        std=CIFAR10_STD,
        is_bcos=False,
    ):
        trans = [
            transforms.RandomHorizontalFlip(),
            transforms.RandomCrop(32, padding=4),
            transforms.ToTensor(),
            custom_transforms.AddInverse()
            if is_bcos
            else transforms.Normalize(mean=mean, std=std),
        ]

        self.transforms = transforms.Compose(trans)

    def __call__(self, img):
        return self.transforms(img)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.transforms)})"

    def __rich_repr__(self):
        yield "transforms", self.transforms.transforms


class CIFAR10ClassificationPresetTest:
    def __init__(
        self,
        mean=CIFAR10_MEAN,
        std=CIFAR10_STD,
        is_bcos=False,
    ):
        trans = [
            transforms.ToTensor(),
            custom_transforms.AddInverse()
            if is_bcos
            else transforms.Normalize(mean=mean, std=std),
        ]

        self.transforms = transforms.Compose(trans)

    def __call__(self, img):
        return self.transforms(img)

    def __repr__(self):
        return f"{self.__class__.__name__}({repr(self.transforms)})"

    def __rich_repr__(self):
        yield "transforms", self.transforms.transforms


def get_args_dict(ignore: "tuple | list" = tuple()):
    """Helper for saving args easily."""
    import inspect

    frame = inspect.currentframe().f_back
    av = inspect.getargvalues(frame)
    ignore = tuple(ignore) + ("self", "cls")
    return {arg: av.locals[arg] for arg in av.args if arg not in ignore}



"""
Contains categories for the datasets.
"""

__all__ = ["IMAGENET_CATEGORIES", "CIFAR10_CATEGORIES"]

# See https://www.cs.toronto.edu/~kriz/cifar.html
# This is available in torchvision.dataset.CIFAR10 but just to keep stuff consistent...
CIFAR10_CATEGORIES = [
    "airplane",
    "automobile",
    "bird",
    "cat",
    "deer",
    "dog",
    "frog",
    "horse",
    "ship",
    "truck",
]

# Taken from https://github.com/pytorch/vision/blob/11bd2eaa6d6976129836b329b01/torchvision/models/_meta.py
IMAGENET_CATEGORIES = [
    "tench",
    "goldfish",
    "great white shark",
    "tiger shark",
    "hammerhead",
    "electric ray",
    "stingray",
    "cock",
    "hen",
    "ostrich",
    "brambling",
    "goldfinch",
    "house finch",
    "junco",
    "indigo bunting",
    "robin",
    "bulbul",
    "jay",
    "magpie",
    "chickadee",
    "water ouzel",
    "kite",
    "bald eagle",
    "vulture",
    "great grey owl",
    "European fire salamander",
    "common newt",
    "eft",
    "spotted salamander",
    "axolotl",
    "bullfrog",
    "tree frog",
    "tailed frog",
    "loggerhead",
    "leatherback turtle",
    "mud turtle",
    "terrapin",
    "box turtle",
    "banded gecko",
    "common iguana",
    "American chameleon",
    "whiptail",
    "agama",
    "frilled lizard",
    "alligator lizard",
    "Gila monster",
    "green lizard",
    "African chameleon",
    "Komodo dragon",
    "African crocodile",
    "American alligator",
    "triceratops",
    "thunder snake",
    "ringneck snake",
    "hognose snake",
    "green snake",
    "king snake",
    "garter snake",
    "water snake",
    "vine snake",
    "night snake",
    "boa constrictor",
    "rock python",
    "Indian cobra",
    "green mamba",
    "sea snake",
    "horned viper",
    "diamondback",
    "sidewinder",
    "trilobite",
    "harvestman",
    "scorpion",
    "black and gold garden spider",
    "barn spider",
    "garden spider",
    "black widow",
    "tarantula",
    "wolf spider",
    "tick",
    "centipede",
    "black grouse",
    "ptarmigan",
    "ruffed grouse",
    "prairie chicken",
    "peacock",
    "quail",
    "partridge",
    "African grey",
    "macaw",
    "sulphur-crested cockatoo",
    "lorikeet",
    "coucal",
    "bee eater",
    "hornbill",
    "hummingbird",
    "jacamar",
    "toucan",
    "drake",
    "red-breasted merganser",
    "goose",
    "black swan",
    "tusker",
    "echidna",
    "platypus",
    "wallaby",
    "koala",
    "wombat",
    "jellyfish",
    "sea anemone",
    "brain coral",
    "flatworm",
    "nematode",
    "conch",
    "snail",
    "slug",
    "sea slug",
    "chiton",
    "chambered nautilus",
    "Dungeness crab",
    "rock crab",
    "fiddler crab",
    "king crab",
    "American lobster",
    "spiny lobster",
    "crayfish",
    "hermit crab",
    "isopod",
    "white stork",
    "black stork",
    "spoonbill",
    "flamingo",
    "little blue heron",
    "American egret",
    "bittern",
    "crane bird",
    "limpkin",
    "European gallinule",
    "American coot",
    "bustard",
    "ruddy turnstone",
    "red-backed sandpiper",
    "redshank",
    "dowitcher",
    "oystercatcher",
    "pelican",
    "king penguin",
    "albatross",
    "grey whale",
    "killer whale",
    "dugong",
    "sea lion",
    "Chihuahua",
    "Japanese spaniel",
    "Maltese dog",
    "Pekinese",
    "Shih-Tzu",
    "Blenheim spaniel",
    "papillon",
    "toy terrier",
    "Rhodesian ridgeback",
    "Afghan hound",
    "basset",
    "beagle",
    "bloodhound",
    "bluetick",
    "black-and-tan coonhound",
    "Walker hound",
    "English foxhound",
    "redbone",
    "borzoi",
    "Irish wolfhound",
    "Italian greyhound",
    "whippet",
    "Ibizan hound",
    "Norwegian elkhound",
    "otterhound",
    "Saluki",
    "Scottish deerhound",
    "Weimaraner",
    "Staffordshire bullterrier",
    "American Staffordshire terrier",
    "Bedlington terrier",
    "Border terrier",
    "Kerry blue terrier",
    "Irish terrier",
    "Norfolk terrier",
    "Norwich terrier",
    "Yorkshire terrier",
    "wire-haired fox terrier",
    "Lakeland terrier",
    "Sealyham terrier",
    "Airedale",
    "cairn",
    "Australian terrier",
    "Dandie Dinmont",
    "Boston bull",
    "miniature schnauzer",
    "giant schnauzer",
    "standard schnauzer",
    "Scotch terrier",
    "Tibetan terrier",
    "silky terrier",
    "soft-coated wheaten terrier",
    "West Highland white terrier",
    "Lhasa",
    "flat-coated retriever",
    "curly-coated retriever",
    "golden retriever",
    "Labrador retriever",
    "Chesapeake Bay retriever",
    "German short-haired pointer",
    "vizsla",
    "English setter",
    "Irish setter",
    "Gordon setter",
    "Brittany spaniel",
    "clumber",
    "English springer",
    "Welsh springer spaniel",
    "cocker spaniel",
    "Sussex spaniel",
    "Irish water spaniel",
    "kuvasz",
    "schipperke",
    "groenendael",
    "malinois",
    "briard",
    "kelpie",
    "komondor",
    "Old English sheepdog",
    "Shetland sheepdog",
    "collie",
    "Border collie",
    "Bouvier des Flandres",
    "Rottweiler",
    "German shepherd",
    "Doberman",
    "miniature pinscher",
    "Greater Swiss Mountain dog",
    "Bernese mountain dog",
    "Appenzeller",
    "EntleBucher",
    "boxer",
    "bull mastiff",
    "Tibetan mastiff",
    "French bulldog",
    "Great Dane",
    "Saint Bernard",
    "Eskimo dog",
    "malamute",
    "Siberian husky",
    "dalmatian",
    "affenpinscher",
    "basenji",
    "pug",
    "Leonberg",
    "Newfoundland",
    "Great Pyrenees",
    "Samoyed",
    "Pomeranian",
    "chow",
    "keeshond",
    "Brabancon griffon",
    "Pembroke",
    "Cardigan",
    "toy poodle",
    "miniature poodle",
    "standard poodle",
    "Mexican hairless",
    "timber wolf",
    "white wolf",
    "red wolf",
    "coyote",
    "dingo",
    "dhole",
    "African hunting dog",
    "hyena",
    "red fox",
    "kit fox",
    "Arctic fox",
    "grey fox",
    "tabby",
    "tiger cat",
    "Persian cat",
    "Siamese cat",
    "Egyptian cat",
    "cougar",
    "lynx",
    "leopard",
    "snow leopard",
    "jaguar",
    "lion",
    "tiger",
    "cheetah",
    "brown bear",
    "American black bear",
    "ice bear",
    "sloth bear",
    "mongoose",
    "meerkat",
    "tiger beetle",
    "ladybug",
    "ground beetle",
    "long-horned beetle",
    "leaf beetle",
    "dung beetle",
    "rhinoceros beetle",
    "weevil",
    "fly",
    "bee",
    "ant",
    "grasshopper",
    "cricket",
    "walking stick",
    "cockroach",
    "mantis",
    "cicada",
    "leafhopper",
    "lacewing",
    "dragonfly",
    "damselfly",
    "admiral",
    "ringlet",
    "monarch",
    "cabbage butterfly",
    "sulphur butterfly",
    "lycaenid",
    "starfish",
    "sea urchin",
    "sea cucumber",
    "wood rabbit",
    "hare",
    "Angora",
    "hamster",
    "porcupine",
    "fox squirrel",
    "marmot",
    "beaver",
    "guinea pig",
    "sorrel",
    "zebra",
    "hog",
    "wild boar",
    "warthog",
    "hippopotamus",
    "ox",
    "water buffalo",
    "bison",
    "ram",
    "bighorn",
    "ibex",
    "hartebeest",
    "impala",
    "gazelle",
    "Arabian camel",
    "llama",
    "weasel",
    "mink",
    "polecat",
    "black-footed ferret",
    "otter",
    "skunk",
    "badger",
    "armadillo",
    "three-toed sloth",
    "orangutan",
    "gorilla",
    "chimpanzee",
    "gibbon",
    "siamang",
    "guenon",
    "patas",
    "baboon",
    "macaque",
    "langur",
    "colobus",
    "proboscis monkey",
    "marmoset",
    "capuchin",
    "howler monkey",
    "titi",
    "spider monkey",
    "squirrel monkey",
    "Madagascar cat",
    "indri",
    "Indian elephant",
    "African elephant",
    "lesser panda",
    "giant panda",
    "barracouta",
    "eel",
    "coho",
    "rock beauty",
    "anemone fish",
    "sturgeon",
    "gar",
    "lionfish",
    "puffer",
    "abacus",
    "abaya",
    "academic gown",
    "accordion",
    "acoustic guitar",
    "aircraft carrier",
    "airliner",
    "airship",
    "altar",
    "ambulance",
    "amphibian",
    "analog clock",
    "apiary",
    "apron",
    "ashcan",
    "assault rifle",
    "backpack",
    "bakery",
    "balance beam",
    "balloon",
    "ballpoint",
    "Band Aid",
    "banjo",
    "bannister",
    "barbell",
    "barber chair",
    "barbershop",
    "barn",
    "barometer",
    "barrel",
    "barrow",
    "baseball",
    "basketball",
    "bassinet",
    "bassoon",
    "bathing cap",
    "bath towel",
    "bathtub",
    "beach wagon",
    "beacon",
    "beaker",
    "bearskin",
    "beer bottle",
    "beer glass",
    "bell cote",
    "bib",
    "bicycle-built-for-two",
    "bikini",
    "binder",
    "binoculars",
    "birdhouse",
    "boathouse",
    "bobsled",
    "bolo tie",
    "bonnet",
    "bookcase",
    "bookshop",
    "bottlecap",
    "bow",
    "bow tie",
    "brass",
    "brassiere",
    "breakwater",
    "breastplate",
    "broom",
    "bucket",
    "buckle",
    "bulletproof vest",
    "bullet train",
    "butcher shop",
    "cab",
    "caldron",
    "candle",
    "cannon",
    "canoe",
    "can opener",
    "cardigan",
    "car mirror",
    "carousel",
    "carpenter's kit",
    "carton",
    "car wheel",
    "cash machine",
    "cassette",
    "cassette player",
    "castle",
    "catamaran",
    "CD player",
    "cello",
    "cellular telephone",
    "chain",
    "chainlink fence",
    "chain mail",
    "chain saw",
    "chest",
    "chiffonier",
    "chime",
    "china cabinet",
    "Christmas stocking",
    "church",
    "cinema",
    "cleaver",
    "cliff dwelling",
    "cloak",
    "clog",
    "cocktail shaker",
    "coffee mug",
    "coffeepot",
    "coil",
    "combination lock",
    "computer keyboard",
    "confectionery",
    "container ship",
    "convertible",
    "corkscrew",
    "cornet",
    "cowboy boot",
    "cowboy hat",
    "cradle",
    "crane",
    "crash helmet",
    "crate",
    "crib",
    "Crock Pot",
    "croquet ball",
    "crutch",
    "cuirass",
    "dam",
    "desk",
    "desktop computer",
    "dial telephone",
    "diaper",
    "digital clock",
    "digital watch",
    "dining table",
    "dishrag",
    "dishwasher",
    "disk brake",
    "dock",
    "dogsled",
    "dome",
    "doormat",
    "drilling platform",
    "drum",
    "drumstick",
    "dumbbell",
    "Dutch oven",
    "electric fan",
    "electric guitar",
    "electric locomotive",
    "entertainment center",
    "envelope",
    "espresso maker",
    "face powder",
    "feather boa",
    "file",
    "fireboat",
    "fire engine",
    "fire screen",
    "flagpole",
    "flute",
    "folding chair",
    "football helmet",
    "forklift",
    "fountain",
    "fountain pen",
    "four-poster",
    "freight car",
    "French horn",
    "frying pan",
    "fur coat",
    "garbage truck",
    "gasmask",
    "gas pump",
    "goblet",
    "go-kart",
    "golf ball",
    "golfcart",
    "gondola",
    "gong",
    "gown",
    "grand piano",
    "greenhouse",
    "grille",
    "grocery store",
    "guillotine",
    "hair slide",
    "hair spray",
    "half track",
    "hammer",
    "hamper",
    "hand blower",
    "hand-held computer",
    "handkerchief",
    "hard disc",
    "harmonica",
    "harp",
    "harvester",
    "hatchet",
    "holster",
    "home theater",
    "honeycomb",
    "hook",
    "hoopskirt",
    "horizontal bar",
    "horse cart",
    "hourglass",
    "iPod",
    "iron",
    "jack-o'-lantern",
    "jean",
    "jeep",
    "jersey",
    "jigsaw puzzle",
    "jinrikisha",
    "joystick",
    "kimono",
    "knee pad",
    "knot",
    "lab coat",
    "ladle",
    "lampshade",
    "laptop",
    "lawn mower",
    "lens cap",
    "letter opener",
    "library",
    "lifeboat",
    "lighter",
    "limousine",
    "liner",
    "lipstick",
    "Loafer",
    "lotion",
    "loudspeaker",
    "loupe",
    "lumbermill",
    "magnetic compass",
    "mailbag",
    "mailbox",
    "maillot",
    "maillot tank suit",
    "manhole cover",
    "maraca",
    "marimba",
    "mask",
    "matchstick",
    "maypole",
    "maze",
    "measuring cup",
    "medicine chest",
    "megalith",
    "microphone",
    "microwave",
    "military uniform",
    "milk can",
    "minibus",
    "miniskirt",
    "minivan",
    "missile",
    "mitten",
    "mixing bowl",
    "mobile home",
    "Model T",
    "modem",
    "monastery",
    "monitor",
    "moped",
    "mortar",
    "mortarboard",
    "mosque",
    "mosquito net",
    "motor scooter",
    "mountain bike",
    "mountain tent",
    "mouse",
    "mousetrap",
    "moving van",
    "muzzle",
    "nail",
    "neck brace",
    "necklace",
    "nipple",
    "notebook",
    "obelisk",
    "oboe",
    "ocarina",
    "odometer",
    "oil filter",
    "organ",
    "oscilloscope",
    "overskirt",
    "oxcart",
    "oxygen mask",
    "packet",
    "paddle",
    "paddlewheel",
    "padlock",
    "paintbrush",
    "pajama",
    "palace",
    "panpipe",
    "paper towel",
    "parachute",
    "parallel bars",
    "park bench",
    "parking meter",
    "passenger car",
    "patio",
    "pay-phone",
    "pedestal",
    "pencil box",
    "pencil sharpener",
    "perfume",
    "Petri dish",
    "photocopier",
    "pick",
    "pickelhaube",
    "picket fence",
    "pickup",
    "pier",
    "piggy bank",
    "pill bottle",
    "pillow",
    "ping-pong ball",
    "pinwheel",
    "pirate",
    "pitcher",
    "plane",
    "planetarium",
    "plastic bag",
    "plate rack",
    "plow",
    "plunger",
    "Polaroid camera",
    "pole",
    "police van",
    "poncho",
    "pool table",
    "pop bottle",
    "pot",
    "potter's wheel",
    "power drill",
    "prayer rug",
    "printer",
    "prison",
    "projectile",
    "projector",
    "puck",
    "punching bag",
    "purse",
    "quill",
    "quilt",
    "racer",
    "racket",
    "radiator",
    "radio",
    "radio telescope",
    "rain barrel",
    "recreational vehicle",
    "reel",
    "reflex camera",
    "refrigerator",
    "remote control",
    "restaurant",
    "revolver",
    "rifle",
    "rocking chair",
    "rotisserie",
    "rubber eraser",
    "rugby ball",
    "rule",
    "running shoe",
    "safe",
    "safety pin",
    "saltshaker",
    "sandal",
    "sarong",
    "sax",
    "scabbard",
    "scale",
    "school bus",
    "schooner",
    "scoreboard",
    "screen",
    "screw",
    "screwdriver",
    "seat belt",
    "sewing machine",
    "shield",
    "shoe shop",
    "shoji",
    "shopping basket",
    "shopping cart",
    "shovel",
    "shower cap",
    "shower curtain",
    "ski",
    "ski mask",
    "sleeping bag",
    "slide rule",
    "sliding door",
    "slot",
    "snorkel",
    "snowmobile",
    "snowplow",
    "soap dispenser",
    "soccer ball",
    "sock",
    "solar dish",
    "sombrero",
    "soup bowl",
    "space bar",
    "space heater",
    "space shuttle",
    "spatula",
    "speedboat",
    "spider web",
    "spindle",
    "sports car",
    "spotlight",
    "stage",
    "steam locomotive",
    "steel arch bridge",
    "steel drum",
    "stethoscope",
    "stole",
    "stone wall",
    "stopwatch",
    "stove",
    "strainer",
    "streetcar",
    "stretcher",
    "studio couch",
    "stupa",
    "submarine",
    "suit",
    "sundial",
    "sunglass",
    "sunglasses",
    "sunscreen",
    "suspension bridge",
    "swab",
    "sweatshirt",
    "swimming trunks",
    "swing",
    "switch",
    "syringe",
    "table lamp",
    "tank",
    "tape player",
    "teapot",
    "teddy",
    "television",
    "tennis ball",
    "thatch",
    "theater curtain",
    "thimble",
    "thresher",
    "throne",
    "tile roof",
    "toaster",
    "tobacco shop",
    "toilet seat",
    "torch",
    "totem pole",
    "tow truck",
    "toyshop",
    "tractor",
    "trailer truck",
    "tray",
    "trench coat",
    "tricycle",
    "trimaran",
    "tripod",
    "triumphal arch",
    "trolleybus",
    "trombone",
    "tub",
    "turnstile",
    "typewriter keyboard",
    "umbrella",
    "unicycle",
    "upright",
    "vacuum",
    "vase",
    "vault",
    "velvet",
    "vending machine",
    "vestment",
    "viaduct",
    "violin",
    "volleyball",
    "waffle iron",
    "wall clock",
    "wallet",
    "wardrobe",
    "warplane",
    "washbasin",
    "washer",
    "water bottle",
    "water jug",
    "water tower",
    "whiskey jug",
    "whistle",
    "wig",
    "window screen",
    "window shade",
    "Windsor tie",
    "wine bottle",
    "wing",
    "wok",
    "wooden spoon",
    "wool",
    "worm fence",
    "wreck",
    "yawl",
    "yurt",
    "web site",
    "comic book",
    "crossword puzzle",
    "street sign",
    "traffic light",
    "book jacket",
    "menu",
    "plate",
    "guacamole",
    "consomme",
    "hot pot",
    "trifle",
    "ice cream",
    "ice lolly",
    "French loaf",
    "bagel",
    "pretzel",
    "cheeseburger",
    "hotdog",
    "mashed potato",
    "head cabbage",
    "broccoli",
    "cauliflower",
    "zucchini",
    "spaghetti squash",
    "acorn squash",
    "butternut squash",
    "cucumber",
    "artichoke",
    "bell pepper",
    "cardoon",
    "mushroom",
    "Granny Smith",
    "strawberry",
    "orange",
    "lemon",
    "fig",
    "pineapple",
    "banana",
    "jackfruit",
    "custard apple",
    "pomegranate",
    "hay",
    "carbonara",
    "chocolate sauce",
    "dough",
    "meat loaf",
    "pizza",
    "potpie",
    "burrito",
    "red wine",
    "espresso",
    "cup",
    "eggnog",
    "alp",
    "bubble",
    "cliff",
    "coral reef",
    "geyser",
    "lakeside",
    "promontory",
    "sandbar",
    "seashore",
    "valley",
    "volcano",
    "ballplayer",
    "groom",
    "scuba diver",
    "rapeseed",
    "daisy",
    "yellow lady's slipper",
    "corn",
    "acorn",
    "hip",
    "buckeye",
    "coral fungus",
    "agaric",
    "gyromitra",
    "stinkhorn",
    "earthstar",
    "hen-of-the-woods",
    "bolete",
    "ear",
    "toilet tissue",
]



import os
import time
from typing import List

try:
    import pytorch_lightning as pl
    from pytorch_lightning.utilities import rank_zero_info
except ImportError:
    raise ImportError(
        "Please install pytorch-lightning for using data modules: "
        "`pip install pytorch-lightning`"
    )

import torch.utils.data as data
import torchvision
from torchvision.datasets import CIFAR10, ImageFolder

import bcos.settings as settings

from .categories import CIFAR10_CATEGORIES, IMAGENET_CATEGORIES
from .sampler import RASampler
from .transforms import RandomCutmix, RandomMixup, SplitAndGrid

__all__ = ["ImageNetDataModule", "CIFAR10DataModule", "ClassificationDataModule"]


class ClassificationDataModule(pl.LightningDataModule):
    """Base class for data modules for classification tasks."""

    NUM_CLASSES: int = None
    """Number of classes in the dataset."""
    NUM_TRAIN_EXAMPLES: int = None
    """Number of training examples in the dataset. Need not be defined."""
    NUM_EVAL_EXAMPLES: int = None
    """Number of evaluation examples in the dataset. Need not be defined."""
    CATEGORIES: List[str] = None
    """List of categories in the dataset. Need not be defined."""

    # ===================================== [ Registry stuff ] ======================================
    __data_module_registry = {}
    """Registry of data modules."""

    def __init_subclass__(cls, **kwargs):
        # check that the class attributes are defined
        super().__init_subclass__(**kwargs)
        assert cls.NUM_CLASSES is not None
        # rest don't need to be defined

        # get name and remove DataModule suffix
        name = cls.__name__
        # check if name matches XXXDataModule
        if not name.endswith("DataModule"):
            raise ValueError(
                f"Data module class name '{name}' does not end with 'DataModule'"
            )
        name = name[: -len("DataModule")]
        # check if name is already registered
        if name in cls.__data_module_registry:
            raise ValueError(f"Data module {name} already registered")
        # register the class in the registry
        cls.__data_module_registry[name] = cls

    @classmethod
    def registry(cls):
        """Returns the registry of data modules."""
        return cls.__data_module_registry

    # ===================================== [ Normal stuff ] ======================================
    def __init__(self, config):
        super().__init__()

        self.config = config
        self.batch_size = config["batch_size"]
        self.num_workers = config["num_workers"]

        self.train_dataset = None
        self.eval_dataset = None

        mixup_alpha = config.get("mixup_alpha", 0.0)
        cutmix_alpha = config.get("cutmix_alpha", 0.0)
        p_gridified = config.get("p_gridified", 0.0)
        self.train_collate_fn = self.get_train_collate_fn(
            mixup_alpha, cutmix_alpha, p_gridified
        )

    def train_dataloader(self):
        train_sampler = self.get_train_sampler()
        shuffle = None if train_sampler is not None else True
        return data.DataLoader(
            self.train_dataset,
            self.batch_size,
            shuffle=shuffle,
            sampler=train_sampler,
            num_workers=self.num_workers,
            collate_fn=self.train_collate_fn,
            pin_memory=True,
        )

    def val_dataloader(self):
        return data.DataLoader(
            self.eval_dataset,
            self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    def test_dataloader(self):
        return data.DataLoader(
            self.eval_dataset,
            self.batch_size,
            shuffle=False,
            num_workers=self.num_workers,
            pin_memory=True,
        )

    @classmethod
    def get_train_collate_fn(
        cls,
        mixup_alpha: float = 0.0,
        cutmix_alpha: float = 0.0,
        p_gridified: float = 0.0,
    ):
        assert not (p_gridified and mixup_alpha), "For now, do not use both."

        collate_fn = None
        if p_gridified:
            gridify = SplitAndGrid(p_gridified, num_classes=cls.NUM_CLASSES)

            def collate_fn(batch):
                return gridify(*data.default_collate(batch))

            rank_zero_info(f"Gridify active for training with {p_gridified=}")

        mixup_transforms = []
        if mixup_alpha > 0.0:
            mixup_transforms.append(
                RandomMixup(cls.NUM_CLASSES, p=1.0, alpha=mixup_alpha)
            )
            rank_zero_info(f"Mixup active for training with {mixup_alpha=}")
        if cutmix_alpha > 0.0:
            mixup_transforms.append(
                RandomCutmix(cls.NUM_CLASSES, p=1.0, alpha=cutmix_alpha)
            )
            rank_zero_info(f"Cutmix active for training with {cutmix_alpha=}")
        if mixup_transforms:
            mixupcutmix = torchvision.transforms.RandomChoice(mixup_transforms)

            def collate_fn(batch):  # noqa: F811
                return mixupcutmix(*data.default_collate(batch))

        return collate_fn

    def get_train_sampler(self):
        train_sampler = None

        # see https://github.com/Lightning-AI/lightning/blob/612d43e5bf38ba73b4f372d64594c2f9a32e6d6a/src/pytorch_lightning/trainer/connectors/data_connector.py#L336
        # and https://github.com/Lightning-AI/lightning/blob/612d43e5bf38ba73b4f372d64594c2f9a32e6d6a/src/lightning_lite/utilities/seed.py#L54
        seed = int(os.getenv("PL_GLOBAL_SEED", 0))
        ra_reps = self.config.get("ra_repetitions", None)
        if ra_reps is not None:
            rank_zero_info(f"Activating RASampler with {ra_reps=}")
            train_sampler = RASampler(
                self.train_dataset,
                shuffle=True,
                seed=seed,
                repetitions=ra_reps,
            )

        return train_sampler


class ImageNetDataModule(ClassificationDataModule):
    # from https://image-net.org/download.php
    NUM_CLASSES: int = 1000

    NUM_TRAIN_EXAMPLES: int = 1_281_167
    NUM_EVAL_EXAMPLES: int = 50_000

    CATEGORIES: List[str] = IMAGENET_CATEGORIES

    def __init__(self, config):
        super().__init__(config)
        self.prepare_data_per_node = self.config.get("cache_dataset", None) == "shm"

    def prepare_data(self) -> None:
        cache_dataset = self.config.get("cache_dataset", None)
        if cache_dataset != "shm":
            return

        # print because we also want global non-zero rank's
        start = time.perf_counter()
        print("Caching dataset into SHM!...")
        from .caching import cache_tar_files_to_shm

        cache_tar_files_to_shm()
        end = time.perf_counter()
        print(f"Caching successful! Time taken {end - start:.2f}s")

    def setup(self, stage: str) -> None:
        # this way changes to the settings are reflected at function call time
        SHMTMPDIR = settings.SHMTMPDIR
        IMAGENET_PATH = settings.IMAGENET_PATH
        if stage == "fit":
            cache_dataset = self.config.get("cache_dataset", None)
            rank_zero_info("Setting up ImageNet train dataset...")
            start = time.perf_counter()
            train_root = os.path.join(
                SHMTMPDIR if cache_dataset == "shm" else IMAGENET_PATH,
                "train",
            )
            self.train_dataset = ImageFolder(
                root=train_root,
                transform=self.config["train_transform"],
            )
            assert len(self.train_dataset) == self.NUM_TRAIN_EXAMPLES
            rank_zero_info(f"Done! Took time {time.perf_counter() - start:.2f}s")

            if cache_dataset == "onthefly":
                rank_zero_info("Trying to setup Bagua's cached dataset!")
                from .caching import CachedImageFolder

                self.train_dataset = CachedImageFolder(self.train_dataset)
                rank_zero_info("Successfully setup cached dataset!")

        start = time.perf_counter()
        rank_zero_info("Setting up ImageNet val dataset...")
        self.eval_dataset = ImageFolder(
            root=os.path.join(IMAGENET_PATH, "val"),
            transform=self.config["test_transform"],
        )
        assert len(self.eval_dataset) == self.NUM_EVAL_EXAMPLES
        rank_zero_info(f"Done! Took time {time.perf_counter() - start:.2f}s")


class CIFAR10DataModule(ClassificationDataModule):
    # from https://www.cs.toronto.edu/~kriz/cifar.html
    NUM_CLASSES: int = 10

    NUM_TRAIN_EXAMPLES: int = 50_000
    NUM_EVAL_EXAMPLES: int = 10_000

    CATEGORIES: List[str] = CIFAR10_CATEGORIES

    def setup(self, stage: str) -> None:
        DATA_ROOT = settings.DATA_ROOT
        if stage == "fit":
            self.train_dataset = CIFAR10(
                root=DATA_ROOT,
                train=True,
                transform=self.config["train_transform"],
                download=True,
            )
            assert len(self.train_dataset) == self.NUM_TRAIN_EXAMPLES

        self.eval_dataset = CIFAR10(
            root=DATA_ROOT,
            train=False,
            transform=self.config["test_transform"],
            download=True,
        )
        assert len(self.eval_dataset) == self.NUM_EVAL_EXAMPLES



"""
Nothing fancy, just for creating optimizers.

Modified from: https://github.com/pytorch/vision/blob/0504df5ddf9431909130e7788faf054/references/classification/train.py
"""
import copy
import warnings
from typing import List, Optional, Tuple, Union

import torch

__all__ = ["OptimizerFactory"]


class OptimizerFactory:
    def __init__(
        self,
        name: str,
        lr: Union[int, float],
        **kwargs,
    ):
        self.name = name
        self.args = dict(
            lr=lr,
            **kwargs,
        )

        assert name.lower() in ["adamw", "adam", "sgd", "rmsprop"]

    def create(self, model: torch.nn.Module) -> torch.optim.Optimizer:
        """
        Creates an optimizer with the preset configuration.
        Parameters
        ----------
        model: torch.nn.Module
            The model to optimize.

        Returns
        -------

        """
        name = self.name.lower()
        args = self.args.copy()

        if "weight_decay" in self.args or name == "adamw":
            warnings.warn(
                "Just FYI: Setting a different WD for "
                "**uncentered** norms is not configured! "
                "You can safely ignore this for centered ones though."
            )
            parameters = self.set_weight_decay(model, self.args["weight_decay"])
        else:
            parameters = model.parameters()

        if name == "sgd":
            optimizer = torch.optim.SGD(
                parameters,
                **args,
            )
        elif name == "rmsprop":
            args["eps"] = args.get("eps", 0.0316)
            args["alpha"] = args.get("alpha", 0.9)
            optimizer = torch.optim.RMSprop(
                parameters,
                **args,
            )
        elif name == "adamw":
            optimizer = torch.optim.AdamW(parameters, **args)
        elif name == "adam":
            if args.get("weight_decay", 0) != 0:
                warnings.warn(
                    "You probably should be using AdamW instead of Adam for WD!"
                )
            optimizer = torch.optim.Adam(parameters, **args)
        else:
            raise ValueError(
                f"Invalid optimizer '{name}'. Only SGD, RMSprop, Adam and AdamW are supported."
            )

        return optimizer

    # with methods
    def with_name(self, new_name):
        """
        Creates a new optimizer factory with same parameters
        except with the new name.
        """
        args = copy.deepcopy(self.args)
        lr = args.pop("lr")
        new_optimizer_factory = type(self)(
            name=new_name,
            lr=lr,
            **args,
        )

        return new_optimizer_factory

    def with_lr(self, lr):
        """
        Creates a new optimizer factory with same parameters
        except with the new lr.
        """
        args = copy.deepcopy(self.args)
        del args["lr"]
        new_optimizer_factory = type(self)(
            name=self.name,
            lr=lr,
            **args,
        )

        return new_optimizer_factory

    def with_args(self, **kwargs):
        assert "lr" not in kwargs, "lr should not be in given args, use with_lr instead"
        assert (
            "name" not in kwargs
        ), "name should not be in given args, use with_name instead"

        new_optimizer_factory = type(self)(
            name=self.name,
            lr=self.args["lr"],
            **kwargs,
        )

        return new_optimizer_factory

    def __repr__(self):
        s = self.__class__.__name__
        s += f"(optimizer='{self.name}'"
        if self.args:
            for k, v in self.args.items():
                s += f", {k}={v}"
        s += ")"

        return s

    def __to_config__(self):
        """See bcos.experiments.utils.sanitize_config for details."""
        return dict(
            name=self.name.lower(),
            **self.args,
        )

    @staticmethod
    def set_weight_decay(
        model: torch.nn.Module,
        weight_decay: float,
        norm_weight_decay: Optional[float] = None,
        norm_classes: Optional[List[type]] = None,
        custom_keys_weight_decay: Optional[List[Tuple[str, float]]] = None,
    ):
        """
        From torchvision reference.
        """
        if not norm_classes:
            norm_classes = [
                torch.nn.modules.batchnorm._BatchNorm,
                torch.nn.LayerNorm,
                torch.nn.GroupNorm,
                torch.nn.modules.instancenorm._InstanceNorm,
                torch.nn.LocalResponseNorm,
            ]
        norm_classes = tuple(norm_classes)

        params = {
            "other": [],
            "norm": [],
        }
        params_weight_decay = {
            "other": weight_decay,
            "norm": norm_weight_decay,
        }
        custom_keys = []
        if custom_keys_weight_decay is not None:
            for key, weight_decay in custom_keys_weight_decay:
                params[key] = []
                params_weight_decay[key] = weight_decay
                custom_keys.append(key)

        def _add_params(module, prefix=""):
            for name, p in module.named_parameters(recurse=False):
                if not p.requires_grad:
                    continue
                is_custom_key = False
                for key in custom_keys:
                    target_name = (
                        f"{prefix}.{name}" if prefix != "" and "." in key else name
                    )
                    if key == target_name:
                        params[key].append(p)
                        is_custom_key = True
                        break
                if not is_custom_key:
                    if norm_weight_decay is not None and isinstance(
                        module, norm_classes
                    ):
                        params["norm"].append(p)
                    else:
                        params["other"].append(p)

            for child_name, child_module in module.named_children():
                child_prefix = f"{prefix}.{child_name}" if prefix != "" else child_name
                _add_params(child_module, prefix=child_prefix)

        _add_params(model)

        param_groups = []
        for key in params:
            if len(params[key]) > 0:
                param_groups.append(
                    {"params": params[key], "weight_decay": params_weight_decay[key]}
                )
        return param_groups



from functools import partial

import bcos.models.convnext as convnext
import bcos.models.densenet as densenet
import bcos.models.resnet as resnet
import bcos.models.vgg as vgg
from bcos.modules.bcosconv2d import BcosConv2d

__all__ = ["get_model"]


def get_arch_builder(arch_name: str):
    arch_builder = None

    if arch_name.startswith("resne"):
        arch_builder = getattr(resnet, arch_name)
    elif arch_name.startswith("densenet"):
        arch_builder = getattr(densenet, arch_name)
    elif arch_name.startswith("vgg"):
        arch_builder = getattr(vgg, arch_name)
    elif arch_name.startswith("convnext"):
        arch_builder = getattr(convnext, arch_name)

    assert arch_builder is not None
    return arch_builder


def get_model(model_config):
    assert model_config.get("is_bcos", False), "Should be true!"
    # extract args
    arch_name = model_config["name"]
    args = model_config["args"]
    bcos_args = model_config["bcos_args"]

    # specify conv layer
    if "conv_layer" not in args:  # for compatibility with hubconf overrides
        args["conv_layer"] = partial(BcosConv2d, **bcos_args)

    assert "norm_layer" in args, "norm_layer is required!"

    # create model
    model = get_arch_builder(arch_name)(**args)

    return model



import math  # noqa

from bcos.data.presets import (
    ImageNetClassificationPresetEval,
    ImageNetClassificationPresetTrain,
)
from bcos.experiments.utils import configs_cli, update_config
from bcos.modules import norms
from bcos.modules.losses import (
    BinaryCrossEntropyLoss,
    UniformOffLabelsBCEWithLogitsLoss,
)
from bcos.optim import LRSchedulerFactory, OptimizerFactory

__all__ = ["CONFIGS"]

NUM_CLASSES = 1_000

# These are mainly based on the recipes from
# https://github.com/pytorch/vision/blob/93723b481d1f6e/references/classification/README.md
DEFAULT_BATCH_SIZE = 128  # * 8 GPUs = 1024 effective
DEFAULT_NUM_EPOCHS = 600
DEFAULT_LR = 1e-3
DEFAULT_CROP_SIZE = 176

DEFAULT_NORM_LAYER = norms.NoBias(norms.BatchNormUncentered2d)  # bnu-linear

DEFAULT_OPTIMIZER = OptimizerFactory(name="Adam", lr=DEFAULT_LR)
DEFAULT_LR_SCHEDULE = LRSchedulerFactory(
    name="cosineannealinglr",
    epochs=DEFAULT_NUM_EPOCHS,
    warmup_method="linear",
    warmup_epochs=5,
    warmup_decay=0.01,
)

DEFAULTS = dict(
    data=dict(
        train_transform=ImageNetClassificationPresetTrain(
            crop_size=DEFAULT_CROP_SIZE,
            auto_augment_policy="ta_wide",
            random_erase_prob=0.1,
            is_bcos=True,
        ),
        test_transform=ImageNetClassificationPresetEval(
            crop_size=224,
            resize_size=232,
            is_bcos=True,
        ),
        batch_size=DEFAULT_BATCH_SIZE,
        num_workers=16,
        num_classes=NUM_CLASSES,
        mixup_alpha=0.2,
        cutmix_alpha=1.0,
        ra_repetitions=4,
    ),
    model=dict(
        is_bcos=True,
        # "name": None,
        args=dict(
            num_classes=NUM_CLASSES,
            norm_layer=DEFAULT_NORM_LAYER,
            logit_bias=-math.log(NUM_CLASSES - 1),
        ),
        bcos_args=dict(
            b=2,
            max_out=1,
        ),
    ),
    criterion=UniformOffLabelsBCEWithLogitsLoss(),
    test_criterion=BinaryCrossEntropyLoss(),
    optimizer=DEFAULT_OPTIMIZER,
    lr_scheduler=DEFAULT_LR_SCHEDULE,
    trainer=dict(
        max_epochs=DEFAULT_NUM_EPOCHS,
    ),
    ema=dict(
        steps=32,
        decay=0.99998,
    ),
    use_agc=True,
)


# helper
def update_default(new_config):
    return update_config(DEFAULTS, new_config)


RESNET_DEPTHS = [50, 152]
resnets = {
    f"resnet_{depth}": update_default(
        dict(
            model=dict(
                name=f"resnet{depth}",
                args=dict(
                    stochastic_depth_prob=0.1,
                ),
            ),
        )
    )
    for depth in RESNET_DEPTHS
}

# the densenet is already a very small model (wrt #params),
# so regularization would probably only hurt performance
DENSENET_DEPTHS = [121]
densenets = {
    f"densenet_{depth}": update_default(
        dict(
            data=dict(
                ra_repetitions=None,
            ),
            model=dict(
                name=f"densenet{depth}",
            ),
        )
    )
    for depth in DENSENET_DEPTHS
}

convnext_bnu = {
    f"convnext_{size}_bnu": update_default(
        dict(
            model=dict(
                name=f"convnext_{size}",
                # in the default config, the norm layer is set to BNU
                # stochastic depth activated by default
            )
        )
    )
    for size in ["tiny", "base"]
}

convnext_pn = {
    f"convnext_{size}_pn": update_default(
        dict(
            model=dict(
                name=f"convnext_{size}",
                args=dict(
                    norm_layer=None,  # use model default pos norm
                ),
            )
        )
    )
    for size in ["tiny", "base"]
}


def change_configs_reduce_batch_size(configs, *names):
    for name in names:
        config = configs[name]
        config["trainer"]["accumulate_grad_batches"] = 2
        config["data"]["batch_size"] //= 2


# -------------------------------------------------------------------------

CONFIGS = dict()
CONFIGS.update(resnets)
CONFIGS.update(densenets)
CONFIGS.update(convnext_bnu)
CONFIGS.update(convnext_pn)

change_configs_reduce_batch_size(
    CONFIGS, "convnext_base_pn", "convnext_base_bnu", "resnet_152"
)


if __name__ == "__main__":
    configs_cli(CONFIGS)



"""
Centered Norms.

Code partially taken from
https://github.com/pytorch/pytorch/blob/9e81c0c3f46a36333e82b799b4afa79b44b6bb59/torch/nn/modules/batchnorm.py

Position Norm implementation:
https://github.com/facebookresearch/ConvNeXt/blob/048efcea897d999aed302f2639b6270aedf8d4c8/models/convnext.py

Positional Normalization:
https://github.com/Boyiliee/Positional-Normalization

"""
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from bcos.modules.common import DetachableModule

__all__ = [
    "AllNorm2d",
    "BatchNorm2d",
    "DetachableGroupNorm2d",
    "DetachableGNInstanceNorm2d",
    "DetachableGNLayerNorm2d",
    "DetachableLayerNorm",
    "DetachablePositionNorm2d",
]


# The easiest way is to use BN3D
class AllNorm2d(nn.BatchNorm3d):
    """
    The AllNorm.
    """

    def __init__(
        self,
        num_features: int,
        *args,
        **kwargs,
    ) -> None:
        # since we do it over the whole thing we have to set
        # this to one
        super().__init__(
            1,
            *args,
            **kwargs,
        )

    def forward(self, input: "Tensor") -> "Tensor":
        original_shape = input.shape
        # (B,C,H,W) -> (B,1,C,H,W)
        input = input.unsqueeze(1)

        # (B,1,C,H,W) normed
        output = super().forward(input)

        # (B,C,H,W) normed
        return output.reshape(original_shape)

    def set_explanation_mode(self, activate: bool = True):
        if activate:
            assert (
                not self.training
            ), "Centered AllNorm only supports explanation mode during .eval()!"


# just for a warnable version
class BatchNorm2d(nn.BatchNorm2d):
    def set_explanation_mode(self, activate: bool = True):
        if activate:
            assert (
                not self.training
            ), "Centered BN only supports explanation mode during .eval()!"


class DetachableGroupNorm2d(nn.GroupNorm, DetachableModule):
    def __init__(self, *args, **kwargs):
        DetachableModule.__init__(self)
        super().__init__(*args, **kwargs)

    def forward(self, input: "Tensor") -> "Tensor":
        # input validation
        assert input.dim() == 4, f"Expected 4D input got {input.dim()}D instead!"
        assert input.shape[1] % self.num_groups == 0, (
            "Number of channels in input should be divisible by num_groups, "
            f"but got input of shape {input.shape} and num_groups={self.num_groups}"
        )

        # use faster version if possible
        if not self.detach:
            return F.group_norm(
                input, self.num_groups, self.weight, self.bias, self.eps
            )

        # ------------ manual GN forward pass -------------
        # separate the groups
        # (N, C, *) -> (N, G, C // G, *)
        N, C = input.shape[:2]
        x = input.reshape(N, self.num_groups, C // self.num_groups, *input.shape[2:])

        # calc stats
        var, mean = torch.var_mean(
            x, dim=tuple(range(2, x.dim())), unbiased=False, keepdim=True
        )
        var = var.detach()
        std = (var + self.eps).sqrt()

        # normalize
        x = (x - mean) / std

        # reshape back
        x = x.reshape(input.shape)

        # affine transformation
        if self.weight is not None:
            x = self.weight[None, ..., None, None] * x

        if self.bias is not None:
            x = x + self.bias[None, ..., None, None]

        return x


class DetachableGNInstanceNorm2d(DetachableGroupNorm2d):
    def __init__(self, num_channels: int, *args, **kwargs):
        super().__init__(
            num_groups=num_channels,
            num_channels=num_channels,
            *args,
            **kwargs,
        )


class DetachableGNLayerNorm2d(DetachableGroupNorm2d):
    """
    A CNN detachable layer norm.
    """

    def __init__(self, num_channels: int, *args, **kwargs):
        super().__init__(
            num_groups=1,
            num_channels=num_channels,
            *args,
            **kwargs,
        )


class DetachableLayerNorm(nn.LayerNorm, DetachableModule):
    """
    A non-CNN detachable Layer Norm.
    This is used for the transformers!
    """

    def __init__(self, *args, **kwargs):
        DetachableModule.__init__(self)
        super().__init__(*args, **kwargs)

    def forward(self, input: "Tensor") -> "Tensor":
        # if not detaching -> just use normal pytorch forward pass
        if not self.detach:
            return F.layer_norm(
                input, self.normalized_shape, self.weight, self.bias, self.eps
            )

        # ------------ manual LN detached forward pass -------------
        d_num = len(self.normalized_shape)

        # calc stats
        var, mean = torch.var_mean(
            input, dim=tuple(range(-d_num, 0)), unbiased=False, keepdim=True
        )
        var = var.detach()
        std = (var + self.eps).sqrt_()

        # normalize
        x = (input - mean) / std

        # affine transformation
        if self.weight is not None:
            x = self.weight * x

        if self.bias is not None:
            x = x + self.bias

        return x


class DetachablePositionNorm2d(nn.LayerNorm, DetachableModule):
    def __init__(
        self,
        features: int,
        eps: float = 1e-5,
        affine: bool = True,
        device=None,
        dtype=None,
    ) -> None:
        assert isinstance(
            features, int
        ), f"Provide #features as an int not {type(features)=}"
        DetachableModule.__init__(self)
        super().__init__(
            normalized_shape=features,
            eps=eps,
            elementwise_affine=affine,
            device=device,
            dtype=dtype,
        )
        self.features = features

    def forward(self, x: Tensor) -> Tensor:
        assert x.dim() == 4, f"input should be 4D not {x.dim()}D"

        # use faster version if possible
        if not self.detach:
            x = x.permute(0, 2, 3, 1)
            x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
            x = x.permute(0, 3, 1, 2)
            return x

        # ------------ manual PN detached forward pass -------------
        # get stats
        var, mean = torch.var_mean(x, dim=1, unbiased=False, keepdim=True)
        var = var.detach()
        std = (var + self.eps).sqrt()

        # normalize
        x = (x - mean) / std

        # affine transformation
        if self.weight is not None:
            x = self.weight[None, ..., None, None] * x
        if self.bias is not None:
            x = x + self.bias[None, ..., None, None]

        return x



"""
Group norm without centering.
Also contains GN-based LayerNorm and InstanceNorm wrappers,
`GNLayerNorm` and `GNInstanceNorm`.
"""
from typing import Optional

import torch.nn as nn
from torch import Tensor

from bcos.modules.common import DetachableModule

__all__ = [
    "group_norm_uncentered",
    "GroupNormUncentered2d",
    "GNInstanceNormUncentered2d",
    "GNLayerNormUncentered2d",
]


def group_norm_uncentered(
    input: Tensor,
    num_groups: int,
    weight: Optional[Tensor] = None,
    bias: Optional[Tensor] = None,
    eps: float = 1e-5,
    detach: bool = False,
):
    """
    Un-centered group norm on batched input
    """
    assert input.shape[1] % num_groups == 0, (
        "Number of channels in input should be divisible by num_groups, "
        f"but got input of shape {input.shape} and num_groups={num_groups}"
    )

    # separate the groups
    # (N, C, *) -> (N, G, C // G, *)
    N, C = input.shape[:2]
    x = input.reshape(N, num_groups, C // num_groups, *input.shape[2:])

    # calc stats
    var = (x.detach() if detach else x).var(
        dim=tuple(range(2, x.dim())), unbiased=False, keepdim=True
    )
    std = (var + eps).sqrt()

    # normalize
    x = x / std

    # reshape back
    x = x.reshape(input.shape)

    # affine transformation
    if weight is not None:
        x = weight[None, ..., None, None] * x

    if bias is not None:
        x = x + bias[None, ..., None, None]

    return x


class GroupNormUncentered2d(nn.GroupNorm, DetachableModule):
    def __init__(
        self, num_groups: int, num_channels: int, eps: float = 1e-5, affine: bool = True
    ) -> None:
        DetachableModule.__init__(self)
        super().__init__(num_groups, num_channels, eps, affine)

    def forward(self, input: Tensor) -> Tensor:
        return group_norm_uncentered(
            input,
            self.num_groups,
            self.weight,
            self.bias,
            self.eps,
            detach=self.detach,
        )


class GNInstanceNormUncentered2d(GroupNormUncentered2d):
    def __init__(self, num_channels: int, *args, **kwargs):
        super().__init__(
            num_groups=num_channels,
            num_channels=num_channels,
            *args,
            **kwargs,
        )


class GNLayerNormUncentered2d(GroupNormUncentered2d):
    def __init__(self, num_channels: int, *args, **kwargs):
        super().__init__(
            num_groups=1,
            num_channels=num_channels,
            *args,
            **kwargs,
        )



# From https://github.com/BaguaSys/bagua
from collections import defaultdict
from typing import Dict, List, Optional, Union

__all__ = ["Store", "ClusterStore"]


class Store:
    """
    Base class for key-value store implementations. Entries are added to store with :meth:`set` or :meth:`mset`, and retrieved
    with :meth:`get` or :meth:`mget`.
    """

    def set(self, key: str, value: Union[str, bytes]):
        """Set a key-value pair."""
        pass

    def get(self, key: str) -> Optional[Union[str, bytes]]:
        """Returns the value associated with :attr:`key`, or ``None`` if the key doesn't exist."""
        pass  # type: ignore

    def num_keys(self) -> int:
        """Returns the number of keys in the current store."""
        pass  # type: ignore

    def clear(self):
        """Delete all keys in the current store."""
        pass

    def mset(self, dictionary: Dict[str, Union[str, bytes]]):
        """
        Set multiple entries at once with a dictionary. Each key-value pair in the :attr:`dictionary` will be set.
        """
        pass

    def mget(self, keys: List[str]) -> List[Optional[Union[str, bytes]]]:
        """
        Retrieve each key's corresponding value and return them in a list with the same order as :attr:`keys`.
        """

        pass  # type: ignore

    def status(self) -> bool:
        """
        Returns ``True`` if the current store is alive.
        """
        pass  # type: ignore

    def shutdown(self):
        """
        Shutdown the managed store instances. Unmanaged instances will not be killed.
        """
        pass


class ClusterStore(Store):
    """
    Base class for distributed key-value stores.

    In cluster store, entries will be sharded equally among multiple store instances based on their keys.

    Args:
        stores(List[Store]): A list of stores to shard entries on.

    """

    def __init__(self, stores: List[Store]):
        self.stores = stores
        self.num_stores = len(stores)

        import xxhash

        def xxh64(x):
            return xxhash.xxh64(x).intdigest()

        self.hash_fn = xxh64

    def _hash_key(self, key: str) -> int:
        hash_code = self.hash_fn(key.encode())
        return hash_code % self.num_stores

    def route(self, key: str) -> Store:
        return (
            self.stores[self._hash_key(key)] if self.num_stores > 1 else self.stores[0]
        )

    def set(self, key: str, value: Union[str, bytes]):
        if self.num_stores == 1:
            return self.stores[0].set(key, value)

        self.route(key).set(key, value)

    def get(self, key: str) -> Optional[Union[str, bytes]]:
        if self.num_stores == 1:
            return self.stores[0].get(key)

        return self.route(key).get(key)

    def num_keys(self) -> int:
        return sum([store.num_keys() for store in self.stores])

    def clear(self):
        for store in self.stores:
            store.clear()

    def mset(self, dictionary: Dict[str, Union[str, bytes]]):
        if self.num_stores == 1:
            return self.stores[0].mset(dictionary)

        route_table = {}
        for k, v in dictionary.items():
            sid = self._hash_key(k)
            m = route_table.get(sid, defaultdict(dict))
            m[k] = v
            route_table[sid] = m

        for sid, m in route_table.items():
            self.stores[sid].mset(m)

    def mget(self, keys: List[str]) -> List[Optional[Union[str, bytes]]]:
        if self.num_stores == 1:
            return self.stores[0].mget(keys)

        route_table = {}
        for k in keys:
            sid = self._hash_key(k)
            ll = route_table.get(sid, [])
            ll.append(k)
            route_table[sid] = ll

        result_map = {}
        for sid, ll in route_table.items():
            ret = self.stores[sid].mget(ll)
            m = {k: v for k, v in zip(ll, ret)}
            result_map = {**result_map, **m}

        return list(map(lambda x: result_map.get(x, None), keys))

    def status(self) -> bool:
        return all([store.status() for store in self.stores])

    def shutdown(self):
        for store in self.stores:
            store.shutdown()



from pathlib import Path
from typing import Literal, Tuple, Union

import numpy as np
import torch

from ..exceptions import EMANotFound, MetricsNotFoundError

__all__ = [
    "Metrics",
]

PathLike = Union[str, Path]


class Metrics(dict):
    """
    A dictionary for storing metrics with some additional helper methods.
    """

    VALIDATION_KEY = "eval_acc1"
    """The key for the validation accuracy."""
    EMA_VALIDATION_KEY = "eval_acc1_ema"
    """The key for the EMA validation accuracy."""

    def __init__(self, other: dict):
        metrics = {name: torch.tensor(value) for name, value in other.items()}
        super().__init__(metrics)

    @classmethod
    def from_metrics_dir(cls, metrics_dir: PathLike) -> "Metrics":
        """
        Loads metrics from a directory.

        Parameters
        ----------
        metrics_dir : PathLike
            The metrics directory to load metrics from.

        Returns
        -------
        Metrics
            The loaded metrics.
        """
        metrics_dir = Path(metrics_dir)
        if not metrics_dir.exists():
            raise MetricsNotFoundError(
                f"Metrics directory '{metrics_dir}' does not exist!"
            )
        metrics = {}
        for metric_file in metrics_dir.glob("*.gz"):
            metric_name = metric_file.stem
            metric_values = np.loadtxt(metric_file)
            metrics[metric_name] = metric_values
        return cls(metrics)

    @classmethod
    def from_experiment_dir(cls, exp_dir: PathLike) -> "Metrics":
        """
        Loads metrics from an experiment directory.

        Parameters
        ----------
        exp_dir : PathLike
            The experiment directory to load metrics from.

        Returns
        -------
        Metrics
            The loaded metrics.
        """
        exp_dir = Path(exp_dir)
        if not exp_dir.exists():
            raise MetricsNotFoundError(
                f"Experiment directory '{exp_dir}' does not exist!"
            )

        metrics_dir = exp_dir / "metrics"
        metrics = cls.from_metrics_dir(metrics_dir)
        return metrics

    def get_best_epoch_and_accuracy(self) -> Tuple[int, float]:
        """
        Gets the best epoch and accuracy w.r.t. the top-1 validation accuracy.

        Returns
        -------
        Tuple[int, float]
            The best epoch and accuracy, respectively.
        """
        return self.find_best_epoch_and_metric_value_for(self.VALIDATION_KEY)

    def get_best_epoch_and_accuracy_ema(self) -> Tuple[int, float]:
        """
        Gets the best EMA epoch and accuracy w.r.t. the top-1 validation accuracy.

        Returns
        -------
        Tuple[int, float]
            The best EMA epoch and accuracy, respectively.
        """
        if self.EMA_VALIDATION_KEY not in self:
            raise EMANotFound("EMA metrics not found!")

        return self.find_best_epoch_and_metric_value_for(self.EMA_VALIDATION_KEY)

    def find_best_epoch_and_metric_value_for(
        self,
        metric_key: str,
        mode: Literal["min", "max"] = "max",
    ) -> Tuple[int, float]:
        """
        Finds the best epoch and metric value in metric collection
        for given metric key/name.

        Parameters
        ----------
        metric_key : str
            The metric key/name to search in.
        mode : Literal["min", "max"]
            Whether to min. or max. to find best. Default: max.

        Returns
        -------
        Tuple[int, float]
            The epoch and best metric value, respectively.
        """

        try:
            metric_values = self[metric_key][:, 1]
        except KeyError:
            if metric_key == self.EMA_VALIDATION_KEY:
                raise EMANotFound()
            else:
                raise
        if mode == "max":
            best_epoch_idx = metric_values.argmax()
        elif mode == "min":
            best_epoch_idx = metric_values.argmin()
        else:
            raise ValueError(f"Unknown {mode=}")

        best_entry = self[metric_key][best_epoch_idx]
        best_epoch = int(best_entry[0])
        best_metric_value = float(best_entry[1])
        return best_epoch, best_metric_value



"""
Adaptive Gradient Clipping (https://arxiv.org/abs/2102.06171)

Code references:
official: https://github.com/deepmind/deepmind-research/tree/master/nfnets
timm: https://github.com/rwightman/pytorch-image-models/blob/18ec173f95aa220af753358bf860b16b6/timm/utils/agc.py
lucidrains: https://gist.github.com/lucidrains/0d6560077edac419ab5d3aa29e674d5c
"""
import torch


def unitwise_norm(x: torch.Tensor, norm_type: float = 2.0) -> torch.Tensor:
    if x.squeeze().ndim <= 1:
        dim = None
        keepdim = False
    elif x.ndim in (2, 3):
        dim = 1
        keepdim = True
    elif x.ndim == 4:  # OIHW
        dim = (1, 2, 3)
        keepdim = True
    else:
        raise ValueError(f"Expected 1 <= x.ndim <= 4. Got {x.ndim=}")

    return x.norm(norm_type, dim=dim, keepdim=keepdim)


def adaptive_clip_grad_(parameters, clip_factor=0.01, eps=1e-3, norm_type=2.0):
    if isinstance(parameters, torch.Tensor):
        parameters = [parameters]
    for p in parameters:
        if p.grad is None:
            continue
        p_data = p.detach()
        g_data = p.grad.detach()
        max_norm = (
            unitwise_norm(p_data, norm_type=norm_type).clamp_(min=eps).mul_(clip_factor)
        )
        grad_norm = unitwise_norm(g_data, norm_type=norm_type)
        clipped_grad = g_data * (max_norm / grad_norm.clamp(min=1e-6))
        new_grads = torch.where(grad_norm < max_norm, g_data, clipped_grad)
        p.grad.detach().copy_(new_grads)



from torchvision import models

__all__ = ["get_model"]


def get_model(model_config):
    assert not model_config.get("is_bcos", False), "Should be false!"

    # extract args
    arch_name = model_config["name"]
    args = model_config["args"]

    # create model
    model = getattr(models, arch_name)(**args)

    return model



"""
This module contains the main public API of the bcos package.
"""
# This module contains the preset transforms for ImageNet and C10
import bcos.data.presets as presets

# This contains single module level transforms. Among others, the AddInverse transform
# used to add inverse RGB channels to the input image.
import bcos.data.transforms as transforms

# This module contains the main models and pretrained models
import bcos.models as models

# This module contains the pretrained model entrypoints
import bcos.models.pretrained as pretrained

# This module contains the main nn.Modules of the package to build custom B-cos models
import bcos.modules as modules

# This module contains some factories to build optimizers and lr schedulers
import bcos.optim as optim

# This module contains the settings of the package
import bcos.settings as settings

# This module contains some utility functions/classes related to B-cos
from bcos.common import (
    BcosUtilMixin,
    explanation_mode,
    gradient_to_image,
    plot_contribution_map,
)
from bcos.version import __version__

__all__ = [
    "presets",
    "transforms",
    "models",
    "pretrained",
    "modules",
    "optim",
    "settings",
    "BcosUtilMixin",
    "explanation_mode",
    "gradient_to_image",
    "plot_contribution_map",
]



"""
Contains a Linear layer which uses the B-cos transform.

NOTE: In case you're wondering why the convolution models do not use
`BcosLinear`, it's because maintaining two versions of essentially
the same thing would be very error-prone during development and testing!
"""
from typing import Union

import torch.linalg as LA
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from .common import DetachableModule

__all__ = ["NormedLinear", "BcosLinear"]


class NormedLinear(nn.Linear):
    """
    Standard linear transform, but with unit norm weights.
    """

    def forward(self, input: Tensor) -> Tensor:
        w = self.weight / LA.vector_norm(self.weight, dim=1, keepdim=True)
        return F.linear(input, w, self.bias)


class BcosLinear(DetachableModule):
    """
    BcosLinear is a linear transform with unit norm weights and a cosine similarity
    activation function. The cosine similarity is calculated between the input
    vector and the weight vector. The output is then scaled by the cosine
    similarity.

    See the paper for more details: https://arxiv.org/abs/2205.10268

    Parameters
    ----------
    in_features : int
        Number of input features
    out_features : int
        Number of output features
    bias : bool
        This is ignored. BcosLinear does not support bias.
    device : Optional[torch.device]
        The device of the weights.
    dtype : Optional[torch.dtype]
        The dtype of the weights.
    b : int | float
        The base of the exponential used to scale the cosine similarity.
    max_out : int
        The number of output vectors to use. If this is greater than 1, the
        output is calculated as the maximum of `max_out` vectors. This is
        equivalent to using a MaxOut activation function.
    """

    def __init__(
        self,
        in_features: int,
        out_features: int,
        bias: bool = False,
        device=None,
        dtype=None,
        b: Union[int, float] = 2,
        max_out: int = 1,
    ) -> None:
        assert not bias
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.bias = False

        self.b = b
        self.max_out = max_out

        self.linear = NormedLinear(
            in_features,
            out_features * self.max_out,
            bias=False,
            device=device,
            dtype=dtype,
        )

    def forward(self, in_tensor: Tensor) -> Tensor:
        """
        Forward pass.
        Args:
            in_tensor: Input tensor. Expected shape: (*, H_in)

        Returns:
            B-cos Linear output on the input tensor.
            Shape: (*, H_out)
        """
        # Simple linear layer
        out = self.linear(in_tensor)

        # MaxOut computation
        if self.max_out > 1:
            M = self.max_out
            O = self.out_features  # noqa: E741
            out = out.unflatten(dim=-1, sizes=(O, M))
            out = out.max(dim=-1, keepdim=False).values

        # if B=1, no further calculation necessary
        if self.b == 1:
            return out

        # Calculating the norm of input vectors ||x||
        norm = LA.vector_norm(in_tensor, dim=-1, keepdim=True) + 1e-12

        # Calculate the dynamic scale (|cos|^(B-1))
        # Note that cos = (x·ŵ)/||x||
        maybe_detached_out = out
        if self.detach:
            maybe_detached_out = out.detach()
            norm = norm.detach()

        if self.b == 2:
            dynamic_scaling = maybe_detached_out.abs() / norm
        else:
            abs_cos = (maybe_detached_out / norm).abs() + 1e-6
            dynamic_scaling = abs_cos.pow(self.b - 1)

        # put everything together
        out = dynamic_scaling * out  # |cos|^(B-1) (ŵ·x)
        return out

    def extra_repr(self) -> str:
        # rest in self.linear
        s = "B={b}"

        if self.max_out > 1:
            s += ", max_out={max_out}"

        # final comma as self.linear is shown in next line
        s += ","

        return s.format(**self.__dict__)



import math
from typing import Tuple

import torch
from torch import Tensor
from torchvision.transforms import functional as F
from torchvision.utils import make_grid


class AddInverse(torch.nn.Module):
    """To a [B, C, H, W] input add the inverse channels of the given one to it.
    Results in a [B, 2C, H, W] output. Single image [C, H, W] is also accepted.

    Args:
        dim (int): where to add channels to. Default: -3
    """

    def __init__(self, dim: int = -3):
        super().__init__()
        self.dim = dim

    def forward(self, in_tensor: Tensor) -> Tensor:
        return torch.cat([in_tensor, 1 - in_tensor], dim=self.dim)


class SplitAndGrid(torch.nn.Module):
    def __init__(self, p_gridified, num_classes):
        super().__init__()
        self.p = p_gridified
        self.num_classes = num_classes

    @staticmethod
    def doIt(imgs):
        return torch.cat(
            list(
                map(
                    lambda x: torch.nn.functional.interpolate(
                        make_grid(x, nrow=2, padding=0)[None],
                        (imgs.shape[-2], imgs.shape[-1]),
                        mode="bilinear",
                    ),
                    imgs.chunk(len(imgs) // 4),
                )
            )
        )

    @staticmethod
    def adjust_targets(target):
        return torch.cat(
            list(
                map(
                    lambda t: t.sum(0, keepdim=True).clamp(0, 1),
                    target.chunk(len(target) // 4),
                )
            )
        )

    def forward(self, batch: Tensor, target: Tensor) -> Tuple[Tensor, Tensor]:
        n = int(len(batch) * self.p)
        gridded = SplitAndGrid.doIt(batch[:n])
        target = torch.nn.functional.one_hot(target, num_classes=self.num_classes).to(
            dtype=batch.dtype
        )
        target = torch.cat([SplitAndGrid.adjust_targets(target[:n]), target[n:]])
        return torch.cat([gridded, batch[n:]], dim=0), target


class RandomMixup(torch.nn.Module):
    """Randomly apply Mixup to the provided batch and targets.
    The class implements the data augmentations as described in the paper
    `"mixup: Beyond Empirical Risk Minimization" <https://arxiv.org/abs/1710.09412>`_.

    Args:
        num_classes (int): number of classes used for one-hot encoding.
        p (float): probability of the batch being transformed. Default value is 0.5.
        alpha (float): hyperparameter of the Beta distribution used for mixup.
            Default value is 1.0.
        inplace (bool): boolean to make this transform inplace. Default set to False.
    """

    def __init__(
        self,
        num_classes: int,
        p: float = 0.5,
        alpha: float = 1.0,
        inplace: bool = False,
    ) -> None:
        super().__init__()

        if num_classes < 1:
            raise ValueError(
                f"Please provide a valid positive value for the num_classes. Got num_classes={num_classes}"
            )

        if alpha <= 0:
            raise ValueError("Alpha param can't be zero.")

        self.num_classes = num_classes
        self.p = p
        self.alpha = alpha
        self.inplace = inplace

    def forward(self, batch: Tensor, target: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Args:
            batch (Tensor): Float tensor of size (B, C, H, W)
            target (Tensor): Integer tensor of size (B, )

        Returns:
            Tensor: Randomly transformed batch.
        """
        if batch.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {batch.ndim}")
        if target.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target.ndim}")
        if not batch.is_floating_point():
            raise TypeError(f"Batch dtype should be a float tensor. Got {batch.dtype}.")
        if target.dtype != torch.int64:
            raise TypeError(f"Target dtype should be torch.int64. Got {target.dtype}")

        if not self.inplace:
            batch = batch.clone()
            target = target.clone()

        if target.ndim == 1:
            target = torch.nn.functional.one_hot(
                target, num_classes=self.num_classes
            ).to(dtype=batch.dtype)

        if torch.rand(1).item() >= self.p:
            return batch, target

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = batch.roll(1, 0)
        target_rolled = target.roll(1, 0)

        # Implemented as on mixup paper, page 3.
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        batch_rolled.mul_(1.0 - lambda_param)
        batch.mul_(lambda_param).add_(batch_rolled)

        target_rolled.mul_(1.0 - lambda_param)
        target.mul_(lambda_param).add_(target_rolled)

        return batch, target

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"num_classes={self.num_classes}"
            f", p={self.p}"
            f", alpha={self.alpha}"
            f", inplace={self.inplace}"
            f")"
        )
        return s


class RandomCutmix(torch.nn.Module):
    """Randomly apply Cutmix to the provided batch and targets.
    The class implements the data augmentations as described in the paper
    `"CutMix: Regularization Strategy to Train Strong Classifiers with Localizable Features"
    <https://arxiv.org/abs/1905.04899>`_.

    Args:
        num_classes (int): number of classes used for one-hot encoding.
        p (float): probability of the batch being transformed. Default value is 0.5.
        alpha (float): hyperparameter of the Beta distribution used for cutmix.
            Default value is 1.0.
        inplace (bool): boolean to make this transform inplace. Default set to False.
    """

    def __init__(
        self,
        num_classes: int,
        p: float = 0.5,
        alpha: float = 1.0,
        inplace: bool = False,
    ) -> None:
        super().__init__()
        if num_classes < 1:
            raise ValueError(
                "Please provide a valid positive value for the num_classes."
            )
        if alpha <= 0:
            raise ValueError("Alpha param can't be zero.")

        self.num_classes = num_classes
        self.p = p
        self.alpha = alpha
        self.inplace = inplace

    def forward(self, batch: Tensor, target: Tensor) -> Tuple[Tensor, Tensor]:
        """
        Args:
            batch (Tensor): Float tensor of size (B, C, H, W)
            target (Tensor): Integer tensor of size (B, )

        Returns:
            Tensor: Randomly transformed batch.
        """
        if batch.ndim != 4:
            raise ValueError(f"Batch ndim should be 4. Got {batch.ndim}")
        if target.ndim != 1:
            raise ValueError(f"Target ndim should be 1. Got {target.ndim}")
        if not batch.is_floating_point():
            raise TypeError(f"Batch dtype should be a float tensor. Got {batch.dtype}.")
        if target.dtype != torch.int64:
            raise TypeError(f"Target dtype should be torch.int64. Got {target.dtype}")

        if not self.inplace:
            batch = batch.clone()
            target = target.clone()

        if target.ndim == 1:
            target = torch.nn.functional.one_hot(
                target, num_classes=self.num_classes
            ).to(dtype=batch.dtype)

        if torch.rand(1).item() >= self.p:
            return batch, target

        # It's faster to roll the batch by one instead of shuffling it to create image pairs
        batch_rolled = batch.roll(1, 0)
        target_rolled = target.roll(1, 0)

        # Implemented as on cutmix paper, page 12 (with minor corrections on typos).
        lambda_param = float(
            torch._sample_dirichlet(torch.tensor([self.alpha, self.alpha]))[0]
        )
        _, H, W = F.get_dimensions(batch)

        r_x = torch.randint(W, (1,))
        r_y = torch.randint(H, (1,))

        r = 0.5 * math.sqrt(1.0 - lambda_param)
        r_w_half = int(r * W)
        r_h_half = int(r * H)

        x1 = int(torch.clamp(r_x - r_w_half, min=0))
        y1 = int(torch.clamp(r_y - r_h_half, min=0))
        x2 = int(torch.clamp(r_x + r_w_half, max=W))
        y2 = int(torch.clamp(r_y + r_h_half, max=H))

        batch[:, :, y1:y2, x1:x2] = batch_rolled[:, :, y1:y2, x1:x2]
        lambda_param = float(1.0 - (x2 - x1) * (y2 - y1) / (W * H))

        target_rolled.mul_(1.0 - lambda_param)
        target.mul_(lambda_param).add_(target_rolled)

        return batch, target

    def __repr__(self) -> str:
        s = (
            f"{self.__class__.__name__}("
            f"num_classes={self.num_classes}"
            f", p={self.p}"
            f", alpha={self.alpha}"
            f", inplace={self.inplace}"
            f")"
        )
        return s



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

__all__ = [
    "BinaryCrossEntropyLoss",
    "UniformOffLabelsBCEWithLogitsLoss",
]


class BinaryCrossEntropyLoss(nn.Module):
    """BCE with optional one-hot from dense targets, label smoothing, thresholding
    from https://github.com/rwightman/pytorch-image-models/blob/a520da9b49/timm/loss/binary_cross_entropy.py

    The label smoothing is done as in `torch.nn.CrossEntropyLoss`.
    In other words, the formula from https://arxiv.org/abs/1512.00567 is strictly followed
    even if input targets samples are sparse, unlike in timm.

    Important: Inputs are assumed to be logits. Targets can be either dense or sparse, and in the latter
    they should not be in logit space.
    """

    def __init__(
        self,
        smoothing=0.0,
        target_threshold: Optional[float] = None,
        weight: Optional[torch.Tensor] = None,
        reduction: str = "mean",
        pos_weight: Optional[torch.Tensor] = None,
    ):
        super(BinaryCrossEntropyLoss, self).__init__()
        assert 0.0 <= smoothing < 1.0
        self.smoothing = smoothing
        self.target_threshold = target_threshold
        self.reduction = reduction
        self.register_buffer("weight", weight)
        self.register_buffer("pos_weight", pos_weight)

    def forward(self, x: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        assert x.shape[0] == target.shape[0]

        num_classes = x.shape[-1]

        # convert dense to sparse
        if target.shape != x.shape:
            target = F.one_hot(target, num_classes=num_classes).to(dtype=x.dtype)

        # apply smoothing if needed
        if self.smoothing > 0.0:
            # just like in `torch.nn.CrossEntropyLoss`
            target = target * (1 - self.smoothing) + self.smoothing / num_classes

        # Make target 0, or 1 if threshold set
        if self.target_threshold is not None:
            target = target.gt(self.target_threshold).to(dtype=target.dtype)

        return F.binary_cross_entropy_with_logits(
            x, target, self.weight, pos_weight=self.pos_weight, reduction=self.reduction
        )

    def extra_repr(self) -> str:
        result = f"reduction={self.reduction}, "
        if self.smoothing > 0:
            result += f"smoothing={self.smoothing}, "
        if self.target_threshold is not None:
            result += f"target_threshold={self.target_threshold}, "
        if self.weight is not None:
            result += f"weight={self.weight.shape}, "
        if self.pos_weight is not None:
            result += f"pos_weight={self.pos_weight.shape}, "
        result = result[:-2]
        return result


class UniformOffLabelsBCEWithLogitsLoss(nn.Module):
    """
    BCE loss with off value targets equal to some value.
    If not provided then it is `1/N`, where `N` is the number of classes.
    The on values are set to 1 as normal.

    This is best explained with an example, as follows:

    Examples
    --------
    Let N=5 and our target be t=3. Then t will be mapped to the following:
    `[0.2, 0.2, 0.2, 1.0, 0.2]`.

    If a particular off value is provided instead for example 2e-3 then it's:
    `[2e-3, 2e-3, 2e-3, 1.0, 2e-3]`
    """

    def __init__(self, reduction: str = "mean", off_label: Optional[float] = None):
        super().__init__()
        self.reduction = reduction
        self.off_label = off_label

    def forward(self, x: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        assert x.shape[0] == target.shape[0]

        num_classes = x.shape[-1]
        off_value = self.off_label or (1.0 / num_classes)
        if target.shape != x.shape:
            target = F.one_hot(target, num_classes=num_classes).to(dtype=x.dtype)

        # make off values (0) to at least 1/N
        target = target.clamp(min=off_value)

        return F.binary_cross_entropy_with_logits(x, target, reduction=self.reduction)

    def extra_repr(self) -> str:
        result = f"reduction={self.reduction}, "
        if self.off_label is not None:
            result += f"off_label={self.off_label}, "
        result = result[:-2]
        return result



"""
Utilities for working with configs.
"""
import argparse
import copy
import difflib
import warnings
from importlib import import_module
from typing import Any, Callable, Dict, List, Tuple, Union

import torch.nn

try:
    from rich import print as pprint
except ImportError:
    from pprint import pprint

from .structure_constants import (
    BASE_EXPERIMENTS_DIRECTORY,
    CONFIGS_MODULE,
    CONFIGS_VAR_NAME,
    MODEL_FACTORY_MODULE,
    MODEL_FACTORY_VAR_NAME,
    ROOT,
)

__all__ = [
    "update_config",
    "configs_cli",
    "get_configs_and_model_factory",
    "sanitize_config",
    "ALLOWED_SANITIZED_DATA_TYPES",
    # specific config creators
    "create_configs_with_different_seeds",
]


def update_config(old_config: Dict, new_config: Dict) -> Dict:
    """
    Creates a new updated config. The old config is copied and updated
    with the newer one. This is done recursively.

    Parameters
    ----------
    old_config : Dict
        The old config.
    new_config : Dict
        The new config to create an updated version of the old one.

    Returns
    -------
    Dict
        The updated version of the old config different from the old one.
    """
    result = copy.deepcopy(old_config)
    for k, v in new_config.items():
        # if subconfig update that recursively
        if k in result and isinstance(result[k], dict):
            assert isinstance(
                v, dict
            ), "Trying to overwrite a dict with something in a config!"
            result[k] = update_config(old_config=result[k], new_config=v)
        else:
            result[k] = v

    return result


def configs_cli(configs: Dict, *argv: str) -> None:
    """
    A CLI for working with configs. Primarily used for debugging.

    Parameters
    ----------
    configs : Dict[str, Dict[str, Any]]
        The configs to work with.
    argv: str
        Optionally, provided arguments to parse directly
        (do not use sys.argv).
    """
    parser = argparse.ArgumentParser(
        "Print config information. By default prints number of configs."
    )
    parser.add_argument(
        "-f",
        "--find",
        type=str,
        default=None,
        help="Check if given config is present and print it.",
    )
    parser.add_argument(
        "-s",
        "--to_script",
        action="store_true",
        default=False,
        help="Create a script file with commands for all experiments.",
    )
    parser.add_argument(
        "-a",
        "--print-all",
        action="store_true",
        default=False,
        help="Print all the names of the configs present.",
    )
    argv = None if len(argv) == 0 else argv
    args = parser.parse_args(argv)

    if len(configs) == 0:
        warnings.warn("No configs found. It's empty!")

    if args.to_script:
        cmd_template = (
            "python run_with_submitit.py --gpus $NUMGPUS --nodes $NUMNODES --timeout 4 --dataset ImageNet "
            "--base_network $BASENET --distributed --csv_logger --wandb_logger "
            "--wandb_project $WANDPROJ --experiment_name {exp_name} --amp --cache_dataset shm"
        )
        with open("run_exps.sh", "w") as file:
            for config in configs:
                file.write(cmd_template.format(exp_name=config) + "\n")
        return

    if args.find is not None:
        if args.find in configs:
            print(f"Found '{args.find}'")
            pprint(configs[args.find])
        else:
            print(f"No config named '{args.find}'!")
            maybe_alternative = difflib.get_close_matches(
                args.find, configs.keys(), n=1
            )
            if maybe_alternative:
                print(f"Did you mean '{maybe_alternative[0]}'?")
    elif args.print_all:
        for name in configs.keys():
            print(name)
    else:
        print(f"There are a total of {len(configs)} configs.")


def get_configs_and_model_factory(
    dataset: str, base_network: str
) -> Tuple[Dict, Callable[..., torch.nn.Module]]:
    """
    Gets all the configs and the model factory function for given
    base network.

    Parameters
    ----------
    dataset : str
        The dataset for the model.
    base_network : str
        The base network.

    Returns
    -------
    A tuple of (configs, model_factory).
    """
    base_module_path = ".".join(
        [ROOT, BASE_EXPERIMENTS_DIRECTORY, dataset, base_network]
    )
    model_path = ".".join([base_module_path, MODEL_FACTORY_MODULE])
    configs_path = ".".join([base_module_path, CONFIGS_MODULE])

    try:
        model_module = import_module(model_path)
    except ModuleNotFoundError:
        print(f"Unable to import '{model_path}'")
        raise
    try:
        configs_module = import_module(configs_path)
    except ModuleNotFoundError:
        print(f"Unable to import '{configs_path}'")
        raise

    return getattr(configs_module, CONFIGS_VAR_NAME), getattr(
        model_module, MODEL_FACTORY_VAR_NAME
    )


ALLOWED_SANITIZED_DATA_TYPES = (str, int, float, bool, tuple, list, type(None))
SANITIZED_DICT_TYPE = Dict[
    str, Union[str, int, float, bool, tuple, list, type(None), "SANITIZED_DICT_TYPE"]
]


def sanitize_config(config_dict: Dict) -> SANITIZED_DICT_TYPE:
    """
    This sanitizes the config dict so that the hyperparameters can be nicely
    logged as strings or numbers or other primitive types except dict.
    In particular, with W&B. This works recursively.

    Parameters
    ----------
    config_dict : Dict[str, Any]
        The config to sanitize.

    Notes
    -----
    If a *value* object has the method `__to_config__`, then that will be
    called instead to get a **dictionary**, instead of just using `repr` on it.
    For an example, see the class `OptimizerFactory`.

    Returns
    -------
    SANITIZED_DICT_TYPE
        The sanitized config.
    """
    result = dict()
    for key, value in config_dict.items():
        if isinstance(value, dict):  # subconfig
            value = sanitize_config(value)
        elif isinstance(value, ALLOWED_SANITIZED_DATA_TYPES):  # one of allowed types
            pass  # do nothing already suitable
        elif hasattr(value, "__to_config__"):  # non-dict object's subconfig
            sub_config_dict = value.__to_config__()
            value = sanitize_config(sub_config_dict)
        else:  # just use repr if nothing else possible
            value = repr(value)

        result[key] = value

    return result


# --------------------------------------------------------------------------------
# config creators
# --------------------------------------------------------------------------------
def create_configs_with_different_seeds(
    configs: Dict[str, Dict[str, Any]], seeds: Union[List[int], int]
):
    """
    For each experiment config in given configs, changes the seed to given seed(s).

    Parameters
    ----------
    configs : Dict[str, Dict[str, Any]]
        The name to config dict.
    seeds: List[int] | int
        Alternatively provide multiple seeds.

    Returns
    -------
    Dict[str, Dict[str, Any]]
        Changed configs.
    """
    if isinstance(seeds, int):
        seeds = [seeds]

    result = dict()
    for seed in seeds:
        new_configs = copy.deepcopy(configs)
        new_configs = {f"{key}-{seed=}": value for key, value in new_configs.items()}
        for name, config in new_configs.items():
            config["seed"] = seed
        result.update(new_configs)

    return result



"""
This module contains entrypoints to loading pretrained models.

Notes
-----
Doing it this way allows:
#. a better overview of the available weights
#. less coupling of model definition code and our config system
#. easy integration with torch.hub

There is obviously a tradeoff here, as this requires slightly more code
but given the advantages I think it's worth it.
"""
from typing import Any, Callable, Dict, List

import torch

from bcos.experiments.utils import Experiment

__all__ = ["list_available"]  # entrypoints filled by `register` decorator


_entrypoint_registry = {}


def register(
    entrypoint_fn: Callable[..., torch.nn.Module]
) -> Callable[..., torch.nn.Module]:
    """Decorator to register a function as an entrypoint."""
    _entrypoint_registry[entrypoint_fn.__name__] = entrypoint_fn
    __all__.append(entrypoint_fn.__name__)
    return entrypoint_fn


def list_available() -> List[str]:
    """List all available pretrained models."""
    return [k for k in _entrypoint_registry.keys()]


BASE = "https://github.com/B-cos/B-cos-v2/releases/download/v0.0.1-weights"


# map (base -> (model_name -> url))
URLS: Dict[str, Dict[str, str]] = {
    "bcos_final": {
        # resnets and resnext
        "resnet_18": f"{BASE}/resnet_18-68b4160fff.pth",
        "resnet_34": f"{BASE}/resnet_34-a63425a03e.pth",
        "resnet_50": f"{BASE}/resnet_50-ead259efe4.pth",
        "resnet_101": f"{BASE}/resnet_101-84c3658278.pth",
        "resnet_152": f"{BASE}/resnet_152-42051a77c1.pth",
        "resnext_50_32x4d": f"{BASE}/resnext_50_32x4d-57af241ab9.pth",
        # densenets
        "densenet_121": f"{BASE}/densenet_121-b8daf96afb.pth",
        "densenet_161": f"{BASE}/densenet_161-9e9ea51353.pth",
        "densenet_169": f"{BASE}/densenet_169-7037ee0604.pth",
        "densenet_201": f"{BASE}/densenet_201-00ac87066f.pth",
        # other
        "vgg_11_bnu": f"{BASE}/vgg_11_bnu-34036029f0.pth",
    },
    "bcos_final_long": {
        "convnext_tiny_pn": f"{BASE}/convnext_tiny_pn-539b1bfb37.pth",
        "convnext_base_pn": f"{BASE}/convnext_base_pn-b0495852c6.pth",
        "convnext_tiny_bnu": f"{BASE}/convnext_tiny_bnu-dbd7f5ef9d.pth",
        "convnext_base_bnu": f"{BASE}/convnext_base_bnu-7c32a704b3.pth",
        "densenet_121": f"{BASE}/densenet_121_long-5175461597.pth",
        "resnet_50": f"{BASE}/resnet_50_long-ef38a88533.pth",
        "resnet_152": f"{BASE}/resnet_152_long-0b4b434939.pth",
    },
}


def _get_model(
    experiment_name: str,
    pretrained: bool,
    progress: bool,
    base_network: str = "bcos_final",
    dataset: str = "ImageNet",
    **model_kwargs: Any,
) -> torch.nn.Module:
    """
    Helper that loads the model and attaches its config and
    transform to it as `config` and `transform` respectively.
    """
    # load empty model
    exp = Experiment(dataset, base_network, experiment_name)
    bcos_args = {}
    if "bcos_args" in model_kwargs:
        bcos_args = dict(bcos_args=model_kwargs.pop("bcos_args"))
    model = exp.get_model(args=model_kwargs, **bcos_args)

    # attach stuff
    assert not hasattr(model, "config")
    assert not hasattr(model, "transform")
    model.config = exp.config
    model.transform = model.config["data"]["test_transform"]

    # load weights if needed
    if pretrained:
        url = URLS[base_network][experiment_name]
        state_dict = torch.hub.load_state_dict_from_url(
            url,
            progress=progress,
            check_hash=True,
        )
        model.load_state_dict(state_dict)

    return model


# ------------------------------- [model entrypoints] -------------------------------------
@register
def resnet18(
    pretrained: bool = False, progress: bool = True, **kwargs
) -> "torch.nn.Module":
    r"""B-cos ResNet-18.

    B-cos version of a ResNet-18 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  68.736%      |
    +---------+---------------+
    | Acc@5   |  87.430%      |
    +---------+---------------+
    | #Params |  11.69M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnet_18", pretrained, progress, **kwargs)


@register
def resnet34(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-34

    B-cos version of a ResNet-34 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  72.284%      |
    +---------+---------------+
    | Acc@5   |  90.052%      |
    +---------+---------------+
    | #Params |  21.80M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnet_34", pretrained, progress, **kwargs)


@register
def resnet50(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-50

    B-cos version of a ResNet-50 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  75.882%      |
    +---------+---------------+
    | Acc@5   |  92.064%      |
    +---------+---------------+
    | #Params |  25.54M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnet_50", pretrained, progress, **kwargs)


@register
def resnet101(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-101

    B-cos version of a ResNet-101 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  76.532%      |
    +---------+---------------+
    | Acc@5   |  92.538%      |
    +---------+---------------+
    | #Params |  44.50M       |
    +---------+---------------+

    References
    ----------
    .. [1] `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    .. [2] `Deep Residual Learning for Image Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnet_101", pretrained, progress, **kwargs)


@register
def resnet152(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-152

    B-cos version of a ResNet-152 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  76.484%      |
    +---------+---------------+
    | Acc@5   |  92.398%      |
    +---------+---------------+
    | #Params |  60.13M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/pdf/1512.03385.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnet_152", pretrained, progress, **kwargs)


@register
def resnext50_32x4d(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNeXt-50 32x4d

    B-cos version of a ResNeXt-50 32x4d model.

    +---------+---------------------+
    | Name    | Value               |
    +=========+=====================+
    | Acc@1   |  75.820%            |
    +---------+---------------------+
    | Acc@5   |  91.810%            |
    +---------+---------------------+
    | #Params |  25.00M             |
    +---------+---------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Aggregated Residual Transformations for Deep Neural Networks <https://arxiv.org/pdf/1611.05431.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("resnext_50_32x4d", pretrained, progress, **kwargs)


@register
def densenet121(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos DenseNet-121

    B-cos version of a DenseNet-121 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  73.612%      |
    +---------+---------------+
    | Acc@5   |  91.106%      |
    +---------+---------------+
    | #Params |  7.95M        |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Densely Connected Convolutional Networks <https://arxiv.org/pdf/1608.06993.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("densenet_121", pretrained, progress, **kwargs)


@register
def densenet161(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos DenseNet-161

    B-cos version of a DenseNet-161 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  76.622%      |
    +---------+---------------+
    | Acc@5   |  92.554%      |
    +---------+---------------+
    | #Params |  28.58M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Densely Connected Convolutional Networks <https://arxiv.org/pdf/1608.06993.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("densenet_161", pretrained, progress, **kwargs)


@register
def densenet169(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos DenseNet-169

    B-cos version of a DenseNet-169 model.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  75.186%      |
    +---------+---------------+
    | Acc@5   |  91.786%      |
    +---------+---------------+
    | #Params |  14.08M       |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Densely Connected Convolutional Networks <https://arxiv.org/pdf/1608.06993.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("densenet_169", pretrained, progress, **kwargs)


@register
def densenet201(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos DenseNet-201

    B-cos version of a DenseNet-201 model.

    +---------+-----------------+
    | Name    | Value           |
    +=========+=================+
    | Acc@1   |  75.480%        |
    +---------+-----------------+
    | Acc@5   |  91.992%        |
    +---------+-----------------+
    | #Params |  19.91M         |
    +---------+-----------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Densely Connected Convolutional Networks <https://arxiv.org/pdf/1608.06993.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("densenet_201", pretrained, progress, **kwargs)


@register
def vgg11_bnu(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos VGG-11 BNU

    B-cos version of a VGG-11 model with Batch Normalization without centering.

    +---------+---------------+
    | Name    | Value         |
    +=========+===============+
    | Acc@1   |  69.310%      |
    +---------+---------------+
    | Acc@5   |  88.388%      |
    +---------+---------------+
    | #Params |  132.86M      |
    +---------+---------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Very Deep Convolutional Networks for Large-Scale Image Recognition <https://arxiv.org/pdf/1409.1556.pdf>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model("vgg_11_bnu", pretrained, progress, **kwargs)


# ----------------------------------------------------------------------------------------------------------------------
# Models trained much longer (600 epochs) with better accuracies.
# ----------------------------------------------------------------------------------------------------------------------
@register
def convnext_tiny(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ConvNeXt-T

    B-cos version of a ConvNeXt-Tiny model with HW-normalization.

    The weights are from epoch 596's (starting from 0) EMA checkpoint weights.
    The model was trained with AMP.

    +---------+------------------+
    | Name    | Value            |
    +=========+==================+
    | Acc@1   |  77.488%         |
    +---------+------------------+
    | Acc@5   |  93.192%         |
    +---------+------------------+
    | #Params |  28.54M           |
    +---------+------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `A ConvNet for the 2020s <https://arxiv.org/abs/2201.03545>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "convnext_tiny_pn",
        pretrained,
        progress,
        base_network="bcos_final_long",
        **kwargs,
    )


@register
def convnext_base(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ConvNeXt-B

    B-cos version of a ConvNeXt-Base model with HW-normalization.

    The weights are from epoch 581's (starting from 0) non-EMA checkpoint weights.
    The model was trained with AMP.

    +---------+------------------+
    | Name    | Value            |
    +=========+==================+
    | Acc@1   |  79.650%         |
    +---------+------------------+
    | Acc@5   |  94.614%         |
    +---------+------------------+
    | #Params |  88.47M          |
    +---------+------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `A ConvNet for the 2020s <https://arxiv.org/abs/2201.03545>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "convnext_base_pn",
        pretrained,
        progress,
        base_network="bcos_final_long",
        **kwargs,
    )


@register
def convnext_tiny_bnu(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ConvNeXt-T with BNU

    B-cos version of a ConvNeXt-Tiny model with Batch Normalization without centering.

    The weights are from epoch 456's (starting from 0) EMA checkpoint weights.
    The model was trained with AMP.

    +---------+------------------+
    | Name    | Value            |
    +=========+==================+
    | Acc@1   |  76.826%         |
    +---------+------------------+
    | Acc@5   |  93.090%         |
    +---------+------------------+
    | #Params |  28.54M          |
    +---------+------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `A ConvNet for the 2020s <https://arxiv.org/abs/2201.03545>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "convnext_tiny_bnu",
        pretrained,
        progress,
        base_network="bcos_final_long",
        **kwargs,
    )


@register
def convnext_base_bnu(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ConvNeXt-B with BNU

    B-cos version of a ConvNeXt-Base model with Batch Normalization without centering.

    The weights are from epoch 541's (starting from 0) EMA checkpoint weights.
    The model was trained with AMP.

    +---------+------------------+
    | Name    | Value            |
    +=========+==================+
    | Acc@1   |  80.142%         |
    +---------+------------------+
    | Acc@5   |  94.834%         |
    +---------+------------------+
    | #Params |  88.47M          |
    +---------+------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `A ConvNet for the 2020s <https://arxiv.org/abs/2201.03545>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "convnext_base_bnu",
        pretrained,
        progress,
        base_network="bcos_final_long",
        **kwargs,
    )


@register
def densenet121_long(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos DenseNet-121 just trained longer.

    The model architecture is the same as the one from `densenet121`.

    The weights are from epoch 596's (starting from 0) non-EMA checkpoint weights.
    The model was trained with AMP.

    +---------+---------------------+
    | Name    | Value               |
    +=========+=====================+
    | Acc@1   |  77.302%            |
    +---------+---------------------+
    | Acc@5   |  93.234%            |
    +---------+---------------------+
    | #Params |  7.95M              |
    +---------+---------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Densely Connected Convolutional Networks <https://arxiv.org/abs/1608.06993>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "densenet_121", pretrained, progress, base_network="bcos_final_long", **kwargs
    )


@register
def resnet50_long(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-50 just trained longer.

    The model architecture is the same as the one from `resnet50`.

    The weights are from epoch 580's (starting from 0) non-EMA checkpoint weights.
    The model was trained with AMP.

    +---------+------------------+
    | Name    | Value            |
    +=========+==================+
    | Acc@1   |  79.468%         |
    +---------+------------------+
    | Acc@5   |  94.452%         |
    +---------+------------------+
    | #Params |  25.54M          |
    +---------+------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/abs/1512.03385>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "resnet_50", pretrained, progress, base_network="bcos_final_long", **kwargs
    )


@register
def resnet152_long(pretrained: bool = False, progress: bool = True, **kwargs):
    """B-cos ResNet-152 just trained longer.

    The model architecture is the same as the one from `resnet152`.
    The weights are from epoch 433's (starting from 0) non-EMA checkpoint weights.
    The model was trained with AMP.

    +---------+-------------------+
    | Name    | Value             |
    +=========+===================+
    | Acc@1   |  80.144%          |
    +---------+-------------------+
    | Acc@5   |  94.116%          |
    +---------+-------------------+
    | #Params |  60.13M           |
    +---------+-------------------+

    References
    ----------
    `B-cos Networks: Alignment is All We Need for Interpretability <https://arxiv.org/abs/2205.10268>`_

    `Deep Residual Learning for Image Recognition <https://arxiv.org/abs/1512.03385>`_

    `TorchVision's new recipe
    <https://pytorch.org/blog/how-to-train-state-of-the-art-models-using-torchvision-latest-primitives>`_

    Parameters
    ----------
    pretrained : bool
        If True, returns a model pre-trained on ImageNet
    progress : bool
        If True, displays a progress bar of the download to stderr
    **kwargs : Any, optional
        Additional arguments passed to the model constructor
        Please see source code for details.
    """
    return _get_model(
        "resnet_152", pretrained, progress, base_network="bcos_final_long", **kwargs
    )



"""
For reproducing numbers.
"""
import math  # noqa

import torch.nn as nn

from bcos.data.presets import (
    ImageNetClassificationPresetEval,
    ImageNetClassificationPresetTrain,
)
from bcos.experiments.utils import (
    configs_cli,
    create_configs_with_different_seeds,
    update_config,
)
from bcos.optim import LRSchedulerFactory, OptimizerFactory

__all__ = ["CONFIGS"]

NUM_CLASSES = 1_000

# The recipes have been taken from:
# https://github.com/pytorch/vision/blob/93723b481d1f6e/references/classification/README.md
DEFAULT_BATCH_SIZE = 64  # per GPU! * 4 = 256 effective
DEFAULT_NUM_EPOCHS = 90
DEFAULT_LR = 0.1
DEFAULT_CROP_SIZE = 224

# no norm layer specification because the models set BN as default anyways

DEFAULT_OPTIMIZER = OptimizerFactory(
    name="SGD",
    lr=DEFAULT_LR,
    momentum=0.9,
    weight_decay=1e-4,
)
DEFAULT_LR_SCHEDULE = LRSchedulerFactory(
    name="steplr",
    step_size=30,
    gamma=0.1,
)

DEFAULTS = dict(
    data=dict(
        train_transform=ImageNetClassificationPresetTrain(
            crop_size=DEFAULT_CROP_SIZE,
        ),
        test_transform=ImageNetClassificationPresetEval(
            crop_size=DEFAULT_CROP_SIZE,
        ),
        batch_size=DEFAULT_BATCH_SIZE,
        num_workers=8,
        num_classes=NUM_CLASSES,
    ),
    model=dict(
        is_bcos=False,
        # "name": None,
        args=dict(
            num_classes=NUM_CLASSES,
        ),
    ),
    criterion=nn.CrossEntropyLoss(),
    test_criterion=nn.CrossEntropyLoss(),
    optimizer=DEFAULT_OPTIMIZER,
    lr_scheduler=DEFAULT_LR_SCHEDULE,
    trainer=dict(
        max_epochs=DEFAULT_NUM_EPOCHS,
    ),
)


# helper
def update_default(new_config):
    return update_config(DEFAULTS, new_config)


RESNET_DEPTHS = [18, 34, 50, 101, 152]
resnets = {
    f"resnet_{depth}": update_default(
        dict(
            model=dict(
                name=f"resnet{depth}",
            ),
        )
    )
    for depth in RESNET_DEPTHS
}

DENSENET_DEPTHS = [121, 161, 201, 169]
densenets = {
    f"densenet_{depth}": update_default(
        dict(
            model=dict(
                name=f"densenet{depth}",
            ),
        )
    )
    for depth in DENSENET_DEPTHS
}

vggs = {
    "vgg_11_bn": update_default(
        dict(
            model=dict(
                name="vgg11_bn",
            ),
        )
    ),
}

# -------------------------------------------------------------------------

CONFIGS = dict()
CONFIGS.update(resnets)
CONFIGS.update(densenets)
CONFIGS.update(vggs)
CONFIGS.update(create_configs_with_different_seeds(CONFIGS, seeds=[420, 1337]))


if __name__ == "__main__":
    configs_cli(CONFIGS)
